/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is "Java-Python Extension (JPE)".
 * 
 * The Initial Developer of the Original Code is Frederic Bruno Giacometti.
 * Portions created by Frederic Bruno Giacometti are
 * Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
 * 
 * Contributor(s): frederic.giacometti@arakne.com
 *
 * Acknowledgments:
 * Particular gratitude is expressed to the following parties for their
 * contributing support to the development of JPE:
 *     - Arakne SARL (France) and its staff (http://www.arakne.com);
 *       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
 *     - The European Union through its funding of the IST Alice project
 *     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
 *       coordinator of the IST Alice project
 *     - iPhrase Technologies Inc. (Massachusetts, USA)
 */

#include <jni.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef _WIN32
#include <signal.h>
#endif
#include <Python.h>
#include "jutil.h"
#include "jpe.h"

#include "python_Javapy.h"

static PyObject_t s_format_exception = NULL;
static jobject s_jrlock = NULL;

struct excdesc
{
  jclass class;
  jmethodID init; /*<init> method*/
};

static struct excdesc* excdesc_create( JNIEnv* env, char const* className)
{
  struct excdesc* result;
  jclass cls;
  result = malloc( sizeof *result);
  if (NOT result)
    {
      LogTrace;
      (*env)->FatalError( env, "malloc() failed");
    }
  cls = (*env)->FindClass( env, className);
  if (NOT cls)
    {
      LogTrace;
      free( result);
      return NULL;
    }
  result->class = (*env)->NewGlobalRef( env, cls);
  (*env)->DeleteLocalRef( env, cls);
  result->init = (*env)->GetMethodID( env, result->class, "<init>", "(JJJ)V");
  if (NOT result->init)
    {
      LogTrace;
      (*env)->DeleteGlobalRef( env, cls);
      free( result);
      return NULL;      
    }
  return result;
}

static void excdesc_cleanup( JNIEnv* env, struct excdesc** pexcdesc)
{
  if (NOT *pexcdesc) return;
  (*env)->DeleteGlobalRef( env, (*pexcdesc)->class);
  free( *pexcdesc);
  *pexcdesc = NULL;
}

static struct
{
  int initialized;
  struct excdesc* PyException, *KeyError;
} s_errclasses = { 0, NULL, NULL};

static int s_errclasses_init( JNIEnv* env)
{
  if (s_errclasses.initialized)
    {
#ifdef JPE_DEBUG
      (void)fprintf( stderr, "%s:%i s_errclasses_init(): already initialized\n",
		     __FILE__, __LINE__);
#endif
      return 0;
    }

  jutil_rlock_take( env, s_jrlock);

  if (s_errclasses.initialized) {LogTrace; goto EXIT;}
  
  s_errclasses.PyException = excdesc_create( env, "python/PyException");
  if (NOT s_errclasses.PyException) {LogTrace; goto EXIT;}

  s_errclasses.KeyError = excdesc_create( env, "python/KeyError");
  if (NOT s_errclasses.KeyError) {LogTrace; goto EXIT;}

  s_errclasses.initialized = 1;

 EXIT:
  if (NOT s_errclasses.initialized)
    excdesc_cleanup( env, &s_errclasses.PyException);

  jutil_rlock_release( env, s_jrlock);

  return NOT s_errclasses.initialized;
}

#ifdef XJPE_DEBUG
# define ThRestore( threadState)					\
/* if (NOT threadState) JPEThrowNull( int); */				\
if (threadState) PyEval_AcquireThread( (PyThreadState*)(threadState));	\
if (NOT _PyThreadState_Current)						\
  {									\
    char msg[ 1024];							\
    jclass err;								\
    (void)sprintf( msg, "%s:%i NOT _PyThreadState_Current",		\
		  __FILE__, __LINE__);					\
									\
    err = jutil_local_class( env,"java/lang/IllegalStateException");	\
    jutil_throw( env, err, msg);					\
    (*env)->DeleteLocalRef( env, err);					\
    return 0;								\
  }
#else
# define ThRestore( threadState)					\
if (threadState) PyEval_AcquireThread( (PyThreadState*)(threadState));
#endif


#ifdef XJPE_DEBUG
#define ThRelease( threadState)						\
if (threadState) PyEval_ReleaseThread( (PyThreadState*)(threadState));
#else
#define ThRelease( threadState)						\
if (threadState) PyEval_ReleaseThread( (PyThreadState*)(threadState));
#endif

#define Jthrow_PyErr( env) jthrow_PyErr( env, __FILE__, __LINE__)

static void jthrow_PyErr( JNIEnv* env, string_t file, int lineno)
{
  PyObject_t exctype, exc, traceback;
  jthrowable throw;

  PyErr_Fetch( &exctype, &exc, &traceback);
  if (NOT exc)
    {
      PyErr_Format( PyExc_SystemError,
		    "%s:%i jthrow_PyErr(): no Python exception",
		    file, lineno);
      PyErr_Fetch( &exctype, &exc, &traceback);
      assert( exc);
    }

  if ((*env)->PushLocalFrame( env, 8))
    {
      (void)fprintf( stderr, "%s:%i failed to push local frame\n",
		     __FILE__, __LINE__);
      if ((*env)->ExceptionCheck( env))
	{
	  (*env)->ExceptionDescribe( env);
	  return;
	}
      else (*env)->FatalError( env, __FILE__ "PushLocalFrame failed");
    }

  if (NOT s_errclasses.initialized && s_errclasses_init( env))
    {
      LogTrace;
      if ((*env)->ExceptionCheck( env))(*env)->ExceptionDescribe( env);
      else (*env)->FatalError( env, __FILE__ ": s_errclasses_init() failed");
      return;
    }

  PyErr_NormalizeException( &exctype, &exc, &traceback);
  {
    struct excdesc* edesc;

    edesc = PyErr_GivenExceptionMatches( exc, PyExc_KeyError)
      ? s_errclasses.KeyError : s_errclasses.PyException;
    assert( edesc);
    throw = (*env)->NewObject( env, edesc->class, edesc->init,
			       (jlong)exctype, (jlong)exc, (jlong)traceback);
  }

  if (throw) (*env)->Throw( env, throw);
  else LogTrace;
  (void)(*env)->PopLocalFrame( env, NULL);
  return;
}

#define JPEThrowNull( type)						\
{									\
  jclass err;								\
  char msg[ 1024];							\
  (void)sprintf( msg, "%s:%i", __FILE__, __LINE__);			\
  err = jutil_local_class( env, "java/lang/NullPointerException");	\
  jutil_throw( env, err, msg);						\
  (*env)->DeleteLocalRef( env, err);					\
  return (type)0;							\
}

#define JPEThrowNullVoid						\
{									\
  jclass err;								\
  char msg[ 1024];							\
  (void)sprintf( msg, "%s:%i", __FILE__, __LINE__);			\
  err = jutil_local_class( env, "java/lang/NullPointerException");	\
  jutil_throw( env, err, msg);						\
  (*env)->DeleteLocalRef( env, err);					\
  return;								\
}

/********************************************************/

JNIEXPORT jint JNICALL JNI_OnLoad( JavaVM* vm, void* reserved)
{
#ifdef JPE_DEBUG
  (void)fprintf( stderr, "%s:%i JNI_OnLoad( 0x%x, 0x%x)\n",
		 __FILE__, __LINE__, vm, reserved);
#endif
  return JPE_JNI_VERSION;
}

static void (*s_jpelib_cleanup)( void) = NULL;

JNIEXPORT void JNICALL JNI_OnUnload( JavaVM* vm, void* reserved)
{
#ifdef JPE_DEBUG
  (void)fprintf( stderr, "%s:%i JNI_OnUnLoad( 0x%x, 0x%x)\n",
		 __FILE__, __LINE__, vm, reserved);
#endif
  if (s_jpelib_cleanup) s_jpelib_cleanup();
}

JNIEXPORT void JNICALL Java_python_Javapy_jpe_1init( JNIEnv* env, jclass self,
						     jobject rlock,
						     jobject threadGroup)
{
#ifndef _WIN32
  /* trying to get coredump on Linux after abort() ....? */
  typedef void (* sig_hand_t)( int);
  struct {
    sig_hand_t ILL;
    sig_hand_t ABRT;
    sig_hand_t FPE;
    sig_hand_t BUS;
    sig_hand_t SEGV;
    sig_hand_t SYS;
  } sig;

  /* reset to defaults ? */
   if (SIG_ERR == (sig.ILL  = signal( SIGILL,  SIG_DFL))) assert( 0);
   if (SIG_ERR == (sig.ABRT = signal( SIGABRT, SIG_DFL))) assert( 0);
   if (SIG_ERR == (sig.FPE  = signal( SIGFPE,  SIG_DFL))) assert( 0);
   if (SIG_ERR == (sig.BUS  = signal( SIGBUS,  SIG_DFL))) assert( 0);
   if (SIG_ERR == (sig.SEGV = signal( SIGSEGV, SIG_DFL))) assert( 0);
   /* if (SIG_ERR == (sig.SYS  = signal( SIGSYS,  SIG_DFL))) assert( 0); */
#endif
   s_jrlock = (*env)->NewGlobalRef( env, rlock);
   if (NOT s_jrlock)
     {
       (void)fprintf( stderr, "%s:%i cannot globalref rlock object\n",
		      __FILE__, __LINE__);
       if (NOT (*env)->ExceptionCheck( env))
	 (*env)->FatalError( env, __FILE__ ": stop here");
       return;
     }

   jpe_initTypes();
   jpe_javapyinit( env, threadGroup);
}

JNIEXPORT jboolean JNICALL Java_python_Javapy_Py_1IsInitialized( JNIEnv* env,
							  jclass self)
{
  return (jboolean)Py_IsInitialized();
}

JNIEXPORT jlong JNICALL Java_python_Javapy_Py_1Initialize( JNIEnv* env,
						    jclass self)
{
  if (jpe_Py_Initialize( env)) Jthrow_PyErr( env);
  return (jlong)PyEval_SaveThread();
}

JNIEXPORT void JNICALL Java_python_Javapy_Py_1Finalize( JNIEnv* env,
							jclass self)
{
  Py_Finalize();
}

JNIEXPORT jlong JNICALL Java_python_Javapy_interpFromThread( JNIEnv* env,
							     jclass self,
							     jlong pyThread)
{
  if (NOT pyThread) JPEThrowNull( jlong);
  return (jlong)((PyThreadState*)pyThread)->interp;
}

JNIEXPORT jlong JNICALL Java_python_Javapy_threadStateNew( JNIEnv* env,
							   jclass self,
							   jlong pyInterp)
{
  return (jlong)PyThreadState_New( (PyInterpreterState*)pyInterp);
}

JNIEXPORT void JNICALL Java_python_Javapy_threadStateDelete( JNIEnv* env,
							     jobject self,
							     jlong threadState)
{
  PyEval_AcquireLock();
  PyThreadState_Clear( (PyThreadState*)threadState);
  PyEval_ReleaseLock();
  PyThreadState_Delete( (PyThreadState*)threadState);
}

JNIEXPORT jstring JNICALL Java_python_Javapy_format_1exception( JNIEnv* env,
							 jclass self,
							 jlong threadState,
							 jlong excType,
							 jlong exc,
							 jlong traceback)
{
  PyObject_t ostr;
  jstring result;

  ThRestore( threadState);
  if (NOT s_format_exception)
    {
      PyObject_t throwmod;
      
      throwmod = PyImport_ImportModule( PyModJavaThrowable);
      if (NOT throwmod)
	{
	  (void)fprintf( stderr,
			 "%s:%i format_exception(): "
			 "cannot import %s\n",
			 __FILE__, __LINE__, PyModJavaThrowable);
	  PyErr_Print();
	  {
	    char msg[ 1024];
	    (void)sprintf( msg, "%s:%i Javapy.format_exception()",
			   __FILE__, __LINE__);
	    (*env)-> FatalError( env, msg);
	  }
	}
      s_format_exception = PyObject_GetAttrString( throwmod,
						   "formatException");
      Py_DECREF( throwmod);
      if (NOT s_format_exception)
	{
	  (void)fprintf( stderr,
			 "%s:%i format_exception(): "
			 "cannot find formatException in java.throwable\n",
			 __FILE__, __LINE__);
	  PyErr_Print();
	  (*env)->FatalError( env, "Javapy.format_exception()");
	}
    }
  assert( s_format_exception);
  ostr = PyObject_CallFunction( s_format_exception, "OOO",
				excType ? (PyObject_t)excType : Py_None,
				exc ? (PyObject_t)exc : Py_None,
				traceback ? (PyObject_t)traceback : Py_None);
  if (NOT ostr)
    {
      (void)fprintf( stderr, "%s:%i format_exception() error\n",
		     __FILE__, __LINE__);
      PyErr_Print();
      (*env)->ExceptionDescribe( env);
      (*env)->FatalError( env, "Javapy.format_exception()");
    }
  assert( PyString_Check( ostr));

  result = (*env)->NewStringUTF( env, PyString_AsString( ostr));
  Py_DECREF( ostr);
  ThRelease( threadState);
  return result;
}

/* PyEval */

JNIEXPORT void JNICALL Java_python_Javapy_PyEval_1AcquireThread( JNIEnv* env,
							  jclass self,
							  jlong threadState)
{
  PyEval_AcquireThread( (PyThreadState*)threadState);
}

JNIEXPORT void JNICALL Java_python_Javapy_PyEval_1RestoreThread( JNIEnv* env,
							  jclass self,
							  jlong threadState)
{
  PyEval_RestoreThread( (PyThreadState*)threadState);
}

JNIEXPORT void JNICALL Java_python_Javapy_PyEval_1ReleaseThread( JNIEnv* env,
							  jclass self,
							  jlong threadState)
{
  PyEval_ReleaseThread( (PyThreadState*)threadState);
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyEval_1SaveThread( JNIEnv* env,
							       jclass self)
{
  return (jlong)PyEval_SaveThread();
}

JNIEXPORT void JNICALL Java_python_Javapy_attach( JNIEnv* env,
						  jclass self,
						  jlong threadState,
						  jboolean owned)
{
  if (pyThread_attachJNIEnv( (PyThreadState*)threadState, env, owned, 0))
    {
      LogTrace;
      ThRestore( threadState);
      Jthrow_PyErr( env);
      ThRelease( threadState);
    }
}

/* PyObject */

JNIEXPORT jlong JNICALL Java_python_Javapy_getNone(JNIEnv* env, jclass self,
						   jlong threadState)
{
  jlong result;
  ThRestore( threadState);
  Py_INCREF( Py_None);
  result = (jlong)Py_None;
  ThRelease( threadState);
  return result;
}

JNIEXPORT jlong JNICALL Java_python_Javapy_jobject_1new( JNIEnv* env,
							 jclass self, 
							 jlong threadState,
							 jobject jobj)
{
  PyObject_t pobj;
  ThRestore( threadState);
  pobj = PyJinstance_FromJobject( env, jobj);
  if (NOT pobj) Jthrow_PyErr( env);
  ThRelease( threadState);
  return (jlong)pobj;
}

JNIEXPORT void JNICALL Java_python_Javapy_Py_1DECREF( JNIEnv* env,
						      jclass self,
						      jlong threadState,
						      jlong pobj)
{
  if (NOT pobj) JPEThrowNullVoid;
  ThRestore( threadState);
  Py_DECREF( (PyObject*)pobj);
  if (PyErr_Occurred()) Jthrow_PyErr( env);
  ThRelease( threadState);
}

JNIEXPORT void JNICALL Java_python_Javapy_Py_1INCREF( JNIEnv* env,
					       jclass self,
					       jlong threadState, jlong pobj)
{
  if (NOT pobj) JPEThrowNullVoid;
  ThRestore( threadState);
  Py_INCREF( (PyObject*)pobj);
  if (PyErr_Occurred()) Jthrow_PyErr( env);
  ThRelease( threadState);
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyImport_1ImportModule( JNIEnv* env,
							    jclass self,
							    jlong threadState,
							    jstring name)
{
  PyObject_t result;
  string_t sname;
  sname = (*env)->GetStringUTFChars( env, name, NULL);
  ThRestore( threadState);
  result = PyImport_ImportModule( (char*)sname);
  if (NOT result) Jthrow_PyErr( env);
  ThRelease( threadState);
  (*env)->ReleaseStringUTFChars( env, name, sname);
  return (jlong)result;
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyString_1FromString( JNIEnv* env,
							  jclass self,
							  jlong threadState,
							  jstring name)
{
  PyObject_t result;
  string_t sname;
  sname = (*env)->GetStringUTFChars( env, name, NULL);
  ThRestore( threadState);
  result = PyString_FromString( (char*)sname);
  if (NOT result) Jthrow_PyErr( env);
  ThRelease( threadState);
  (*env)->ReleaseStringUTFChars( env, name, sname);
  return (jlong)result;  
}

JNIEXPORT jobject JNICALL Java_python_Javapy_ptr2object( JNIEnv* env,
							 jclass self,
							 jlong pself)
{
  if (NOT pself) JPEThrowNull( jobject);

  return jobject_fromObject( env, (PyObject_t)pself);
}

JNIEXPORT jint JNICALL Java_python_Javapy_PyObject_1Compare( JNIEnv* env,
							jclass self,
							jlong threadState,
							jlong left,
							jlong right)
{
  int result;
  if (NOT left) JPEThrowNull( jint);
  if (NOT right) JPEThrowNull( jint);
  ThRestore( threadState);
  result = PyObject_Compare( (PyObject_t)left, (PyObject_t)right);
  if (PyErr_Occurred()) Jthrow_PyErr( env);
  ThRelease( threadState);
  return (jint)result;
}

JNIEXPORT jint JNICALL Java_python_Javapy_PyObject_1Hash( JNIEnv* env,
						     jclass self,
						     jlong threadState,
						     jlong obj)
{
  int result;
  if (NOT obj) JPEThrowNull( jint);
  ThRestore( threadState);
  result = PyObject_Hash( (PyObject_t)obj);
  if (PyErr_Occurred()) Jthrow_PyErr( env);
  ThRelease( threadState);
  return (jint)result;
}

JNIEXPORT jstring JNICALL Java_python_Javapy_PyObject_1Str( JNIEnv* env,
						       jclass self,
						       jlong threadState,
						       jlong obj)
{
  jstring result;
  PyObject_t ostring;

  if (NOT obj) JPEThrowNull( jstring);
  ThRestore( threadState);
  ostring = PyObject_Str( (PyObject_t)obj);
  if (NOT ostring) Jthrow_PyErr( env);
  if (PyString_Check( ostring))
    {
      result = (*env)->NewStringUTF( env, PyString_AsString( ostring));
      Py_DECREF( ostring);
    }
  else
    {
      result = NULL;
      PyErr_Format( PyExc_TypeError, "%s:%i PyObject_Str():"
		    "did not return a string", __FILE__, __LINE__);
      Jthrow_PyErr( env);
    }
  ThRelease( threadState);
  return result;
}

JNIEXPORT jstring JNICALL Java_python_Javapy_PyObject_1Repr( JNIEnv* env,
						       jclass self,
						       jlong threadState,
						       jlong obj)
{
  jstring result;
  PyObject_t ostring;

  if (NOT obj) JPEThrowNull( jstring);
  ThRestore( threadState);
  ostring = PyObject_Repr( (PyObject_t)obj);
  if (NOT ostring) Jthrow_PyErr( env);
  if (PyString_Check( ostring))
    {
      result = (*env)->NewStringUTF( env, PyString_AsString( ostring));
      Py_DECREF( ostring);
    }
  else
    {
      result = NULL;
      PyErr_Format( PyExc_TypeError, "%s:%i PyObject_Repr():"
		    "did not return a string", __FILE__, __LINE__);
      Jthrow_PyErr( env);
    }
  ThRelease( threadState);
  return result;
}

JNIEXPORT jobject JNICALL Java_python_Javapy_PyJobject_1AsJobject( JNIEnv* env,
							      jclass self,
							     jlong threadState,
							      jlong obj)
{
  if (NOT obj) JPEThrowNull( jobject);
  ThRestore( threadState);
  PyErr_Format( PyExc_NotImplementedError,
		"%s:%i PyJObject_AsJobject()", __FILE__, __LINE__);
  Jthrow_PyErr( env);  
  ThRelease( threadState);
  return NULL;
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyObject_1GetAttrString( JNIEnv* env,
							     jclass self,
							     jlong threadState,
							     jlong obj,
							     jstring attr)
{
  PyObject_t result;
  string_t sattr;

  if (NOT obj) JPEThrowNull( jlong);
  sattr = (*env)->GetStringUTFChars( env, attr, NULL);
  ThRestore( threadState);
  result = PyObject_GetAttrString( (PyObject_t)obj, (char*)sattr);
  if (NOT result) Jthrow_PyErr( env);
  ThRelease( threadState);
  (*env)->ReleaseStringUTFChars( env, attr, sattr);
  return (jlong)result;
}

JNIEXPORT void JNICALL Java_python_Javapy_PyObject_1SetAttrString( JNIEnv* env,
								   jclass self,
							       jlong threadState,
								   jlong obj,
								   jstring attr,
								   jlong val)
{
  string_t sattr;

  if (NOT obj) JPEThrowNullVoid;
  if (NOT val) JPEThrowNullVoid;
  sattr = (*env)->GetStringUTFChars( env, attr, NULL);
  ThRestore( threadState);
  if (PyObject_SetAttrString( (PyObject_t)obj, (char*)sattr, (PyObject_t)val))
    Jthrow_PyErr( env);
  ThRelease( threadState);
  (*env)->ReleaseStringUTFChars( env, attr, sattr);
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyObject_1CallArray( JNIEnv* env,
							 jclass self,
							 jlong threadState,
							 jlong obj,
							 jobjectArray jarray)
{
  static struct
  {
    jclass class;
    jfieldID pobj;
  } s_PyObject = {NULL, NULL};

  PyObject_t result, oargs;
  jsize argc, idx;

  if (NOT obj) JPEThrowNull( jlong);

  if (NOT s_PyObject.class)
    {
      jclass cls;
      cls = (*env)->FindClass( env, "python/PyObject");
      if (NOT cls) {LogTrace; return 0;}
      s_PyObject.pobj = (*env)->GetFieldID( env, cls, "pself", "J");
      if (NOT s_PyObject.pobj) {LogTrace; return 0;}
      s_PyObject.class = (*env)->NewGlobalRef( env, cls);
      (*env)->DeleteLocalRef( env, cls);
      if (NOT s_PyObject.class) {LogTrace; return 0;}
    }

  argc = (*env)->GetArrayLength( env, jarray);
  ThRestore( threadState);
  oargs = PyTuple_New( argc);
  for (idx = 0; idx < argc; idx++)
    {
      PyObject_t arg;
      jobject jarg;
      jarg = (*env)->GetObjectArrayElement( env, jarray, idx);
      arg = (PyObject_t)(*env)->GetLongField( env, jarg, s_PyObject.pobj);
      Py_INCREF( arg);
      (*env)->DeleteLocalRef( env, jarg);
      PyTuple_SET_ITEM( oargs, idx, arg);
    }
  result = PyObject_CallObject( (PyObject_t)obj, oargs);
  if (NOT result) Jthrow_PyErr( env);
  Py_DECREF( oargs);
  ThRelease( threadState);
  return (jlong)result;  
}

/* PyType */

JNIEXPORT jlong JNICALL Java_python_Javapy_PyType_1Type(JNIEnv* env,
							jclass self,
							jlong threadState)
{
  ThRestore( threadState);
  Py_INCREF( &PyType_Type);
  ThRelease( threadState);
  return (jlong)&PyType_Type;
}

/* PyInstance */

JNIEXPORT jlong JNICALL Java_python_Javapy_PyInstance_1Type( JNIEnv* env,
							     jclass self,
							     jlong threadState)
{
  ThRestore( threadState);
  Py_INCREF( &PyType_Type);
  ThRelease( threadState);
  return (jlong)&PyInstance_Type;
}

/* PyList */

JNIEXPORT jlong JNICALL Java_python_Javapy_PyList_1Type( JNIEnv* env,
							 jclass self,
							 jlong threadState)
{
  ThRestore( threadState);
  Py_INCREF( &PyList_Type);
  ThRelease( threadState);
  return (jlong)&PyList_Type;
}

/* PyMapping */

JNIEXPORT jboolean JNICALL Java_python_Javapy_PyMapping_1Check( JNIEnv* env,
							 jclass self,
							 jlong threadState,
							 jlong pobj)
{
  int result;

 if (NOT pobj) JPEThrowNull( jboolean);
#ifdef JPE_DEBUG
  if (NOT ((PyObject_t)pobj)->ob_type) JPEThrowNull( jboolean);
#endif
  ThRestore( threadState);
  result = PyMapping_Check( (PyObject_t)pobj);
  ThRelease( threadState);
  return (jboolean)result;
}

JNIEXPORT jlong JNICALL Java_python_Javapy_PyMapping_1GetItem( JNIEnv* env,
							       jclass self,
							     jlong threadState,
							       jlong pself,
							       jlong key)
{
  PyObject_t result;

  if (NOT pself) JPEThrowNull( jlong);
  if (NOT key) JPEThrowNull( jlong);
  ThRestore( threadState);
  result = PyObject_GetItem( (PyObject_t)pself, (PyObject_t)key);
  if (NOT result) Jthrow_PyErr( env);
  ThRelease( threadState);
  return (jlong)result;
}

/* PyDict */

JNIEXPORT jlong JNICALL Java_python_Javapy_PyDict_1New( JNIEnv* env,
						        jclass self,
						        jlong threadState)
{
  PyObject_t result;
  ThRestore( threadState);
  result = PyDict_New();
  ThRelease( threadState);
  return (jlong)result;
}
