# 
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
# 
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
# 
# The Original Code is "Java-Python Extension (JPE)".
# 
# The Initial Developer of the Original Code is Frederic Bruno Giacometti.
# Portions created by Frederic Bruno Giacometti are
# Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
# 
# Contributor(s): frederic.giacometti@arakne.com
#
# Acknowledgments:
# Particular gratitude is expressed to the following parties for their
# contributing support to the development of JPE:
#     - Arakne SARL (France) and its staff (http://www.arakne.com);
#       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
#     - The European Union through its funding of the IST Alice project
#     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
#       coordinator of the IST Alice project
#     - iPhrase Technologies Inc. (Massachusetts, USA)
#

"""Java-specific exception classes


Public interface
================

class TypeOverloadingError( TypeError) --
  Thrown when unable to resolve Java overloading, from given Python arguments

class JavaThrowable( Exception, Jobject) -- sibling to java.lang.Throwable

class JavaException( JavaThrowable) -- sibling to java.lang.Exception

class JavaError( JavaThrowable) -- sibling to java.lang.Error

class JavaUnsupportedFlavorException --
  sibling to java.awt.datatransfer.UnsupportedFlavorException


Programming paradigm
====================

All Java exception that can be mapped to an existing Python standard exception
have been defined using double inheritance.
These exception are not published, and are expected to be caught using the
Python standard exception.

Examples:
  * a java.lang.ClassNotFoundException Java exception is caught as a 
    Python ImportError exception.
  * a java.lang.IndexOutOfBoundsException Java exception is caugh as a
    Python IndexError exception

Only Java exception wrappers not fitting in the standard Python
exception hierarchy are public.


Extending exception mapping rules
=================================

The actual map between Java and Python exception classes is defined in
java.throwable.throwMap.
The pairing is performed on the first successful match in climbing up
the hierarchy of the Java throwable object.
The throwMap map can be augmented as needed, statically or dynamically.
"""

__all__ = ('TypeOverloadingError',
           'JavaThrowable', 'JavaException', 'JavaError',
           'JavaUnsupportedFlavorException')

from jobject import Jobject # if import pb, verify that site is properly
                            # imported ...

def formatException( excType, exc, traceback):
    from traceback import format_exception
    from operator import add
    return reduce( add, format_exception( excType, exc, traceback))

class JavaThrowable( Exception, Jobject):
    def __init__( self, jthrowable, msg=''):
        Jobject.__init__( self, jthrowable)
        self.__dict__[ '_msg_'] = msg
    def __str__( self):
        return `(self._jobject_, self._msg_)`
class JavaException( JavaThrowable):
    pass
class JavaError( JavaThrowable):
    pass
class JavaNoSuchFieldError( JavaError, AttributeError):
    pass
class JavaNoSuchFieldException( JavaException, AttributeError):
    pass
class JavaNoSuchMethodException( JavaException, AttributeError):
    pass
class JavaArithmeticException( JavaException, ArithmeticError):
    pass
class JavaIllegalArgumentException( JavaException, TypeError):
    pass
class JavaClassCastException( JavaException, TypeError):
    pass
class JavaIndexOutOfBoundsException( JavaException, IndexError):
    pass
class JavaNoClassDefFoundError( JavaError, ImportError):
    pass
class JavaClassNotFoundException( JavaError, ImportError):
    pass
class JavaNoSuchMethodError( JavaError, AttributeError):
    pass
class JavaIOException( JavaException, IOError):
    pass
class JavaUnsupportedFlavorException( JavaException):
    pass

throwMap = {
    'java.lang.Throwable': JavaThrowable,
    'java.lang.Exception': JavaException,
    'java.lang.Error': JavaError,
    'java.lang.NoSuchFieldError': JavaNoSuchFieldError,
    'java.lang.NoSuchFieldException': JavaNoSuchFieldException,
    'java.lang.NoSuchMethodException': JavaNoSuchMethodException,
    'java.lang.ArithmeticException': JavaArithmeticException,
    'java.lang.IllegalArgumentException': JavaIllegalArgumentException,
    'java.lang.ClassCastException': JavaClassCastException,
    'java.lang.IndexOutOfBoundsException': JavaIndexOutOfBoundsException,
    'java.lang.NoClassDefFoundError': JavaNoClassDefFoundError,
    'java.lang.ClassNotFoundException': JavaClassNotFoundException,
    'java.lang.NoSuchMethodError': JavaNoSuchMethodError,
    'java.io.IOException': JavaIOException,
    'java.awt.datatransfer.UnsupportedFlavorException':
      JavaUnsupportedFlavorException,
    }

def matchThrowable( jobject):
    from string import split
    curclass = jobject.getClass()
    #if __debug__:
        #print 'ww <%s>' % jobject, type( jobject)
        #print 'xx <%s>' % str( curclass)
    while curclass and not throwMap.has_key( str( curclass)):
        curclass = curclass.GetSuperClass()
        #if __debug__: print 'yy <%s>' % str( curclass)
    if curclass:
        return throwMap[ str( curclass)]
    else:
        raise SystemError( jobject, type( jobject),
                           'not a java.lang.Throwable object')

class TypeOverloadingError( TypeError):
    pass
