/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is "Java-Python Extension (JPE)".
 * 
 * The Initial Developer of the Original Code is Frederic Bruno Giacometti.
 * Portions created by Frederic Bruno Giacometti are
 * Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
 * 
 * Contributor(s): frederic.giacometti@arakne.com
 *
 * Acknowledgments:
 * Particular gratitude is expressed to the following parties for their
 * contributing support to the development of JPE:
 *     - Arakne SARL (France) and its staff (http://www.arakne.com);
 *       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
 *     - The European Union through its funding of the IST Alice project
 *     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
 *       coordinator of the IST Alice project
 *     - iPhrase Technologies Inc. (Massachusetts, USA)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Python.h>
#include <jni.h>
#include "jutil.h"
#include "jpe.h"

#ifdef _WIN32
#include <windows.h>
#define dlsym GetProcAddress
#else
#include <dlfcn.h>
#endif

#define JavaEnvKey "JNIEnv"

static PyTypeObject PyJvm_Type, PyJfieldID_Type, PyJmethodID_Type,
  PyJobject_Type;

static jobject s_jrlock = NULL;
static struct jcache* s_jcache = NULL;

static struct jcache* s_jcache_create( JNIEnv* env)
{
  struct jcache* result;

  if (NOT s_jrlock)
    {
      jobject tempref;
      (void)fprintf( stderr, "%s:i Warning s_jcache_create(): "
		     "s_jrlock is NULL; creating it\n", __FILE__, __LINE__);
      tempref = (*env)->NewStringUTF( env, __FILE__ "rlock object");
      if (NOT tempref) return NULL;
      jutil_make_global( env, &tempref);
      if (s_jrlock)
	{ /* not 100% bulletproof */
	  (void)fprintf( stderr, "%s:i Warning s_jcache_create(): RACE LOST\n",
			 __FILE__, __LINE__);
	  
	  (*env)->DeleteGlobalRef( env, tempref);
	}
      else s_jrlock = tempref;
    }

  jutil_rlock_take( env, s_jrlock);
  result = jutil_cache_create( env);
  jutil_rlock_release( env, s_jrlock);
  return result;
}

static struct
{
  PyObject_t matchThrowable, nullException;
} s_jpe = { NULL, NULL};

static PyObject_t jobject_new( JNIEnv* env, jobject obj);
static JNIEnv* pyJavaEnv( PyThreadState* pyThState);
static PyObject_t jfieldID_new( JNIEnv* env,
				int isStatic,
				jclass cls,
				string_t name,
				string_t sig);

static PyObject_t jmethodID_new( JNIEnv* env,
				 int modifier,
				 jclass cls,
				 string_t name,
				 string_t sig);

static const struct jarg Jarg_null = {e_unknown, NULL, NULL};

static PyObject_t jexc2py( JNIEnv* env, string_t filename, int linenum)
{
  static int recurFlag = 0;

  char msg[ 1024];
  jobject thrown;
  PyObject_t jvmException, err, ojthrow;

  if (recurFlag)
    {
      if ((*env)->ExceptionCheck( env)) (*env)->ExceptionDescribe(env );
      PyErr_SetString( PyExc_SystemError, "recursive call to jexc2Py()");
      recurFlag = 0;
      return NULL;
    }
  recurFlag = 1;

  (void)sprintf( msg, "%s:%i Java exception", filename, linenum);
  if (NOT s_jcache)
    {
      (void)fprintf( stderr, "initializations not complete - %s\n", msg);
      (*env)->ExceptionDescribe(env );
    }

  if (NOT (*env)->ExceptionCheck( env))
    {
      if (NOT s_jpe.nullException)
	{
	  PyObject_t throwmod;
	  
	  throwmod = PyImport_ImportModule( PyModJavaThrowable);
	  if (NOT throwmod) return NULL;
	  
	  s_jpe.nullException = PyObject_GetAttrString( throwmod,
							"JavaNullException");
	  Py_DECREF( throwmod);
	  if (NOT s_jpe.nullException) return NULL;
	}
      return PyErr_Format( s_jpe.nullException, msg);
    }

  thrown = (*env)->ExceptionOccurred( env);
  (*env)->ExceptionClear( env);
  if (NOT thrown)
    {
      (*env)->ExceptionDescribe(env );
      PyErr_SetString( PyExc_SystemError, msg);
      recurFlag = 0;
      return NULL;
    }
#ifdef XJPE_DEBUG
  {
    (void)fprintf( stderr, "%s:%i substitute java exception:\n",
		   __FILE__, __LINE__);
    (*env)->CallVoidMethod( env, thrown, s_jcache->Throwable.printStackTrace);
    if ((*env)->ExceptionCheck( env))
      {
	LogTrace;
	(*env)->ExceptionDescribe( env);
      }
  }
#endif
  /* (*env)->ExceptionDescribe( env); */

  ojthrow = jobject_new( env, thrown);
  (*env)->DeleteLocalRef( env, thrown);
  if (NOT ojthrow) return NULL;
  if (NOT s_jpe.matchThrowable)
    {
      PyObject_t throwmod;
      
      throwmod = PyImport_ImportModule( PyModJavaThrowable);
      if (NOT throwmod) return NULL;
      
      s_jpe.matchThrowable = PyObject_GetAttrString( throwmod,
						     "matchThrowable");
      Py_DECREF( throwmod);
      if (NOT s_jpe.matchThrowable) return NULL;
    }
  jvmException = PyObject_CallFunction( s_jpe.matchThrowable, "O", ojthrow);
  if (NOT jvmException)
    {
      Py_DECREF( ojthrow);
      recurFlag = 0;
      return NULL;
    }
  err = PyObject_CallFunction( jvmException, "Os", ojthrow, msg);
  Py_DECREF( ojthrow);
  if (NOT err)
    {
      recurFlag = 0;
      return NULL;
    }
  PyErr_SetObject( jvmException, err);
  Py_DECREF( err);
  recurFlag = 0;
  return NULL;
}

#define Jexc2py( env) (LogTrace, jexc2py( (env), __FILE__, __LINE__))
#define JThreadExc2py( env) (PyEval_RestoreThread( _save), Jexc2py( env))

#define RJexc2py( env) (Jexc2py( env), 1)

#define SJexc2py( env)				\
{						\
    (void)Jexc2py( env);			\
    return 1;					\
}

#define JPE_FRAME_OPEN( env, sz)					\
if ((*(env))->PushLocalFrame( (env), (sz))) return Jexc2py( env);

#define JPE_THFRAME_OPEN( env, sz)					\
if ((*(env))->PushLocalFrame( (env), (sz))) return JThreadExc2py( env);

#define JPE_FRAME_CLOSE( env) ((void)(*(env))->PopLocalFrame( (env), NULL))

#define JPE_ExcFrame2Py( env)\
 (JThreadExc2py( env), JPE_FRAME_CLOSE( env), NULL)

#define ALL_FRAME_OPEN_TYPE( env, sz, retval)				  \
if ((*(env))->PushLocalFrame( (env), (sz))) return Jexc2py( env), retval;

#define ALL_FRAME_OPEN( env, sz) JPE_FRAME_OPEN( env, sz)
#define ALL_FRAME_CLOSE( env) JPE_FRAME_CLOSE( env)

int pyThread_attachJNIEnv( PyThreadState* pyThState, JNIEnv* env,
			   jboolean owned, int hasLock)
{
  PyThreadState* prevThread;
  PyObject_t threadDict;
  int result = 1;

#ifdef JPE_DEBUG
  (void)fprintf( stderr,
		 "%s:%i attachJNIEnv 0x%x owned=%i JNIEnv=0x%x hasLock=%i\n",
		 __FILE__, __LINE__, pyThState, owned, env, hasLock);
#endif

  if (NOT hasLock)
    {
      PyEval_AcquireLock();
      prevThread = PyThreadState_Swap( pyThState);
    }

  threadDict = PyThreadState_GetDict();
  if (NOT threadDict) goto EXIT;

  if (PyDict_GetItemString( threadDict, JavaEnvKey))
    {
      PyObject_t excval;
#ifdef JPE_DEBUG
      (void)fprintf( stderr, "%s:%i, Error: Python thread already attached\n",
		     __FILE__, __LINE__);
#endif
      excval = Py_BuildValue( "Os", PyDict_GetItemString( threadDict,
							  JavaEnvKey),
			      "Python thread already attached");
      if (NOT excval) goto EXIT;
      PyErr_SetObject( PyExc_SystemError, excval);
      Py_DECREF( excval);
      goto EXIT;
    }

  {
    PyObject_t cobj;
    cobj = PyCObject_FromVoidPtr( (void*)env, NULL);
    if (PyDict_SetItemString( threadDict, JavaEnvKey, cobj))
      {
	Py_DECREF( cobj);
	goto EXIT;
      }
  }

  if (NOT s_jcache)
    {
      s_jcache = s_jcache_create( env);
      if (NOT s_jcache)
	{
	  (void)Jexc2py( env);
	  goto EXIT;
	}
    }

  (*env)->CallStaticVoidMethod( env,
				s_jcache->class.PyThreadState,
				s_jcache->PyThreadState.attach,
				(jlong)PyThreadState_Get(),
				owned);
  if ((*env)->ExceptionCheck( env))
    {
      LogTrace;
      (void)Jexc2py( env);
      goto EXIT;
    }

  if (NOT s_jcache)
    {
      s_jcache = s_jcache_create( env);
      if (NOT s_jcache) {LogTrace; assert( 1 == result); goto EXIT;}
    }

  result = 0; /* everything OK */

 EXIT:
  if (NOT hasLock)
    {
      PyThreadState* curThread;
      curThread = PyThreadState_Swap( prevThread);
      assert( curThread == pyThState);
      PyEval_ReleaseLock();
    }
  return result;
}

static JNIEnv* pyThread_getJNIEnv( PyThreadState* pyThread, int hasLock)
{
  JNIEnv* result;
  PyObject_t cenv;
  PyThreadState* prevThread;

  if (NOT pyThread->dict) return NULL;

  if (NOT hasLock)
    {
      PyEval_AcquireLock();
      prevThread = PyThreadState_Swap( pyThread);
    }

  cenv = PyDict_GetItemString( pyThread->dict, JavaEnvKey);
  result = cenv ? (JNIEnv*)PyCObject_AsVoidPtr( cenv) : NULL;

  if (NOT hasLock)
    {
      PyThreadState* curThread;
      curThread = PyThreadState_Swap( prevThread);
      assert( curThread == pyThread);
      PyEval_ReleaseLock();
    }  
  return result;
}

/*** JVM ***/

struct jvm
{
  PyObject_HEAD
  JavaVM* jvm;
  JNIEnv *env;
  string_t libpath;
#ifdef _WIN32
  HINSTANCE jvmlib;
#else
  void* jvmlib;
#endif
};
typedef struct jvm* jvm_t;
#define Py2jvm( obj) ((jvm_t)obj)

static JavaVM* s_JVM = NULL;

static PyObject_t jvm_isInitialized( PyObject_t self, PyObject_t args)
{
  if (NOT PyArg_ParseTuple( args, "")) return NULL;
  return PyInt_FromLong( s_JVM ? 1 : 0);
}

static int javalib_env_initJvm( JNIEnv* env, PyObject_t module)
{
  PyObject_t ojtyp;

  if (NOT s_jcache)
    {
      s_jcache = s_jcache_create( env);
      if (NOT s_jcache) return Jexc2py( env), 1;
    }

#define ModType( lname, sname)					\
  {								\
    ojtyp = jobject_new( env, s_jcache->class.sname);		\
    if (NOT ojtyp) return 1;					\
    if (PyObject_SetAttrString( module, #lname "JType", ojtyp))	\
      return 1;							\
    Py_DECREF( ojtyp);						\
  }

  ModType( Void, V);
  ModType( Boolean, Z);
  ModType( Byte, B);
  ModType( Character, C);
  ModType( Short, S);
  ModType( Integer, I);
  ModType( Long, J);
  ModType( Float, F);
  ModType( Double, D);

  {
    PyObject_t objectPy;
    objectPy = jobject_new( env, s_jcache->class.L);
    if (NOT objectPy) return 1;
    if (PyObject_SetAttrString( module, "ObjectJType", objectPy))
      {Py_DECREF( objectPy); return 1;}
    Py_DECREF( objectPy);
  }

  return 0;
}

static PyObject_t jvm_new( string_t jvmpath, JavaVMInitArgs* vm_args)
{
  jvm_t result;
  JavaVM* jvm;
  JNIEnv* env;
  jint err;
#ifdef _WIN32
  FARPROC creJVM;
  HINSTANCE jvmlib;

  jvmlib = LoadLibrary( jvmpath);
#else
  jint (*creJVM)( JavaVM**, void**, void*);
  void* jvmlib;

  jvmlib = dlopen( jvmpath, RTLD_LAZY);
#endif

#ifdef JPE_DEBUG
  (void)fprintf( stderr, "%s:%i dlopen <%s>\n", __FILE__, __LINE__, jvmpath);
#endif
  if (NOT jvmlib) return PyErr_SetFromErrno( PyExc_OSError);

  creJVM = dlsym( jvmlib, "JNI_CreateJavaVM");
  if (NOT creJVM) return PyErr_SetFromErrno( PyExc_OSError);

  Py_BEGIN_ALLOW_THREADS;
    err = (*creJVM)( &jvm, (void**)&env, vm_args);
  Py_END_ALLOW_THREADS;
  if (err)
    {
      PyErr_SetString( PyExc_SystemError, "JNI_CreateJavaVM() failed");
      return NULL;      
    }
  if (NOT jvm) (void)fprintf( stderr, "WARNING %s:%i jvm_new(): jvm is NULL\n",
			      __FILE__, __LINE__);

  result = PyObject_NEW( struct jvm, &PyJvm_Type);
  result->jvm = jvm;
  result->env = env;
  result->libpath = jvmpath;
  result->jvmlib = jvmlib;

  if (s_JVM)
    {
      Py_DECREF( result);
      return PyErr_Format( PyExc_SystemError, "%s:%i JVM already defined",
			   __FILE__, __LINE__);
    }
  s_JVM = jvm;

  return (PyObject_t)result;
}

static PyObject_t jvm_destroy( PyObject_t self, PyObject_t args)
{
  JavaVM* jvm;
  jint err;
  
  if (args && NOT PyArg_ParseTuple( args, "")) return NULL;
  
  jvm = Py2jvm( self)->jvm;

  if (jvm)
    {
      Py_BEGIN_ALLOW_THREADS;
        err = (*jvm)->DestroyJavaVM( jvm);
      Py_END_ALLOW_THREADS;
      if (err)
	{
	  PyErr_SetString( PyExc_SystemError, "DestroyJavaVM() failed\n");
	  return NULL;
	}
    }

  Py2jvm( self)->jvm = NULL;
  Py2jvm( self)->env = NULL;
  return Py_INCREF( Py_None), Py_None;
}

static PyObject_t jvm_detach( PyObject_t self, PyObject_t args)
{
  JavaVM* jvm;
  jint err;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;
  
  jvm = Py2jvm( self)->jvm;

  Py_BEGIN_ALLOW_THREADS;
    err = (*jvm)->DetachCurrentThread( jvm);
  Py_END_ALLOW_THREADS;
  if (err)
    {
      PyErr_SetString( PyExc_SystemError, "DetachCurrentThread() failed\n");
      return NULL;
    }

  return Py_INCREF( Py_None), Py_None;
}

static void jvm_del( PyObject_t self)
{
  LogTrace;
/*   if (Py2jvm( self)->jvm) {Py_XDECREF( jvm_destroy( self, NULL));} */
}

static struct PyMethodDef jvm_methods[] =
{
  { "destroy", jvm_destroy, METH_VARARGS, "destroy the Java VM"},
  { "detach", jvm_detach, METH_VARARGS, "detach VM from current thread"},
  {NULL, NULL}
};

static JNIEnv* jvm_getjnienv( PyObject_t self)
{
  return Py2jvm( self)->env;
}

static PyObject_t jvm_getattr( PyObject_t self, char* name)
{
  return Py_FindMethod( jvm_methods, self, name);
}

static PyTypeObject PyJvm_Type =
{
  PyObject_HEAD_INIT( NULL)
  0, /*ob_size*/
  "javaVM", /*tp_name*/
  sizeof (struct jvm), /*tb_basicsize*/
  0, /*tp_itemsize*/
  jvm_del, /*tp_dealloc*/
  NULL, /*tp_print*/
  jvm_getattr, /*tp_getattr*/
  NULL, /*tp_setattr*/
  NULL, /*tp_compare*/
  NULL, /*tp_repr*/
  NULL, /*tp_as_number*/
  NULL, /*tp_as_sequence*/
  NULL, /*tp_as_mapping*/
  NULL, /*tp_hash*/
  NULL, /*tp_call*/
  NULL, /*tp_str*/
  NULL, /*tp_getattro*/
  NULL, /*tp_setattro*/
  NULL, /*tp_as_buffer*/
  0, /*tp_xxx4*/
  "Java Virtual Machine", /*tp_doc*/  
};

static PyObject_t jvm_init( PyObject_t self, PyObject_t args)
{
  JavaVMInitArgs vm_args;
  PyObject_t javalib, oargv = NULL, ojvm;
  int idx, sz;
  JavaVMOption* options;
  string_t jvmpath;
  JNIEnv* env = NULL;

  if (NOT PyArg_ParseTuple( args, "O!s|O!",
			    &PyModule_Type, &javalib,
			    &jvmpath,
			    &PyList_Type, &oargv)) return NULL;

  if (oargv)
    {
      sz = PyList_Size( oargv);
      assert( sz);
      options = calloc( 1 + sz, sizeof *options); /* 1+ for safety ?*/
      
      for (idx = 0; idx < sz; idx++)
	{
	  PyObject_t itm;
	  itm = PyList_GetItem( oargv, idx);
	  if (NOT itm) return free( options), NULL;
	  if (NOT PyString_Check( itm))
	    {
	      PyErr_SetString( PyExc_TypeError, "argument is not "
			       "a list of strings");
	      return free( options), NULL;
	    }
	  options[ idx].optionString = PyString_AsString( itm);
	}
    }
  else
    {
      options = NULL;
      sz = 0;
    }

  vm_args.version = JPE_JNI_VERSION;
  vm_args.options = options;
  vm_args.nOptions = sz;
  vm_args.ignoreUnrecognized = JNI_FALSE;

  ojvm = jvm_new( jvmpath, &vm_args);
  free( options);
  if (NOT ojvm) return NULL; 
  if (pyThread_attachJNIEnv( PyThreadState_Get(), jvm_getjnienv( ojvm), 0, 1))
    {
      Py_DECREF( ojvm);
      if (PyErr_Occurred()) return NULL;
      return PyErr_Format( PyExc_SystemError,
			   "%s:%i pyThread_attachJNIEnv failed",
			   __FILE__, __LINE__);
    }

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {Py_DECREF( ojvm); return NULL;}

  if (javalib_env_initJvm( env, javalib)) return NULL;
  {
    PyObject_t javalibmod;

    javalibmod = PyImport_ImportModule( "java.javalib");
    if (NOT javalibmod) { Py_DECREF( ojvm); return NULL; }

    if (javalib_env_initJvm( env, javalibmod))
      { Py_DECREF( javalibmod); Py_DECREF( ojvm); return NULL; }

    Py_DECREF( javalibmod);
  }

  assert( s_jcache);
  (*env)->SetStaticLongField( env,
			      s_jcache->class.PyThreadState,
			      s_jcache->PyThreadState.pyInterp,
			      (jlong)(PyThreadState_Get()->interp));

  return ojvm;
}

static PyObject_t PyString_FromJstring( JNIEnv* env, jstring jstr)
{
  string_t temp;
  PyObject_t result;

  temp = (*env)->GetStringUTFChars( env, jstr, NULL);
  if (NOT temp) return Jexc2py( env);
  result = PyString_FromString( temp);
  (*env)->ReleaseStringUTFChars( env, jstr, temp);
  return result;
}

static jobject s_jThreadGroup = NULL;

void jpe_javapyinit( JNIEnv* env, jobject threadGroup)
{
  assert( NOT s_jThreadGroup);
  s_jThreadGroup = (*env)->NewGlobalRef( env, threadGroup);
  if (NOT s_jThreadGroup)
    {
      (void)fprintf( stderr, "%s:%i cannot globalref threadGroup object\n",
		     __FILE__, __LINE__);
      if (NOT (*env)->ExceptionCheck( env))
	(*env)->FatalError( env, __FILE__ ": stop here");
      return;
    }
}

static JNIEnv* pyJavaEnv( PyThreadState* pyThState)
{
  JNIEnv* jnienv;

  jnienv = pyThread_getJNIEnv( pyThState, 1);

  if (jnienv) return jnienv;

  if (NOT s_JVM)
      return (JNIEnv*)PyErr_Format( PyExc_SystemError,
				    "%s:%i No JVM pointer initialized",
				    __FILE__, __LINE__);
  
  switch ((*s_JVM)->GetEnv( s_JVM, (void**)&jnienv, JPE_JNI_VERSION))
    {
    case JNI_OK:
      break;
    case JNI_EDETACHED:
      { /* attach the thread to Java and get its env pointer */
	static int thCount = 0; /* under Python interpreter lock */
	char thName[ 64];
	struct JavaVMAttachArgs attachargs;

	attachargs.version = JPE_JNI_VERSION;
	(void)sprintf( thName, "attached_from_python_%i", thCount++);
	attachargs.name = thName;
	assert( s_jThreadGroup);
	attachargs.group = s_jThreadGroup;

#ifdef JPE_DEBUG
	(void)fprintf( stderr, "%s:%i attach thread to the JVM\n",
		      __FILE__, __LINE__);
#endif
	if ((*s_JVM)->AttachCurrentThread( s_JVM,
					   (void**)&jnienv, &attachargs))
	  return (JNIEnv*)PyErr_Format( PyExc_SystemError,
					"%s:%i AttachCurrentThread failed",
					__FILE__, __LINE__);
	break;
      }
    case JNI_EVERSION:
      return (JNIEnv*)PyErr_Format( PyExc_SystemError,
				    "%s:%i GetEnv() version error",
				    __FILE__, __LINE__);
    default:
      return (JNIEnv*)PyErr_Format( PyExc_SystemError,
				    "%s:%i GetEnv() error",
				    __FILE__, __LINE__);
    }
      
  if (pyThread_attachJNIEnv( pyThState, jnienv, 0, 1)) return NULL;

  assert( jnienv == pyThread_getJNIEnv( pyThState, 1));
  return jnienv;
}

int jpe_Py_Initialize( JNIEnv* env)
{
  PyEval_InitThreads();
  Py_Initialize();

  if ((*env)->GetJavaVM( env, &s_JVM)) return 1;
  assert( s_JVM);

  return 0;
}

/**** jobject ****/

struct jobject
{
  PyObject_HEAD
  PyObject_t className;
  jobject object;
  enum jtype type, arraytype;
  jsize arraylength;
};
typedef struct jobject* jobject_t;
#define Py2jobject( obj) ((jobject_t)obj)
#define PyJobject_Check( obj) (&PyJobject_Type == (obj)->ob_type)

static int jobject_initType( JNIEnv* env, jobject_t self)
{/* use only Java environment */
  jclass cls;
  string_t className;

  cls = (*env)->GetObjectClass( env, self->object);
  if (NOT cls) {LogTrace; return 1;}
  
  assert( s_jcache);
  className = jutil_class_name( env, s_jcache, cls);
  (*env)->DeleteLocalRef( env, cls);
  if (NOT className) {LogTrace; return 1;}
  
  if (e_jarray == className[ 0])
    {
      self->type = e_jarray;
      assert( e_jarray == className[ 1] || e_jobject == className[ 1]);
      self->arraytype = (enum jtype)className[ 1];
      self->arraylength = (*env)->GetArrayLength( env, self->object);
    }
  else
    {
      Py2jobject( self)->type = e_jobject;
      Py2jobject( self)->arraytype = e_unknown;
    }
  jutil_UTF_cleanup( &className);
  return 0;
}

jobject PyJobject_AsJobject( JNIEnv* env, PyObject_t obj, jarg_t arg)
{
  assert( obj->ob_type == &PyJobject_Type);

  if (arg)
    {
      if (NOT arg->cls)
	{
	  arg->cls = jutil_global_class( env, arg->sig);
	  if (NOT arg->cls) {LogTrace; return NULL;}
	}
      if (NOT (*env)->IsInstanceOf( env, Py2jobject( obj)->object, arg->cls))
	{
	  char msg[ 1024];
	  (void)sprintf( msg, "%s:%i <%s> java object expected",
			 __FILE__, __LINE__, arg->sig);
	  jutil_throw( env, s_jcache->class.ClassCastException, msg);
	  return NULL;
	}
    }

  return (*env)->NewLocalRef( env, Py2jobject( obj)->object);
}

static PyObject_t jobject_new( JNIEnv* env, jobject object)
{
  jobject_t result;
  assert( object);

  result = PyObject_NEW( struct jobject, &PyJobject_Type);
  result->className = NULL;
  result->object = (*env)->NewGlobalRef( env, object);
  result->type = e_unknown;
  result->arraytype = e_unknown;
  result->arraylength = 0;

  return (PyObject_t)result;
}

static void jobject_del( PyObject_t self)
{
  JNIEnv* env;
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return;
  Py_XDECREF( Py2jobject( self)->className);
  (*env)->DeleteGlobalRef( env, Py2jobject( self)->object);  
}

static PyObject_t jobject_string2UTF( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  PyObject_t result;
  string_t str;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}
  ALL_FRAME_OPEN( env, 8);

/*   if (NOT (*env)->IsInstanceOf( env, Py2jobject( self)->object, */
/* 				s_string.class)) */
/*     { */
/*       PyErr_SetString( PyExc_TypeError, "not a java.lang.String object"); */
/*       return NULL; */
/*     } */
  str = (*env)->GetStringUTFChars( env, Py2jobject( self)->object, NULL);
  if (NOT str) return ALL_FRAME_CLOSE( env), Jexc2py( env);
  result = PyString_FromString( str);
  (*env)->ReleaseStringUTFChars( env, Py2jobject( self)->object, str);
  ALL_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_method_Desc( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  PyObject_t result, argtuple;
  jobject selfmeth, parameterTypes;
  jstring returnName, methName;
  int isStatic;
  jsize argc, aidx;
  string_t returnString, methNameStr;
  PyThreadState* _save;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}
  ALL_FRAME_OPEN( env, 8);

  selfmeth = Py2jobject( self)->object;
  if (NOT (*env)->IsInstanceOf( env, selfmeth, s_jcache->class.Method))
    {
      return ALL_FRAME_CLOSE( env),
	  PyErr_Format( PyExc_TypeError,
			"not a java.lang.reflect.Method object");
    }

  Py_UNBLOCK_THREADS;
  JPE_THFRAME_OPEN( env, 32);
  {
    jint modifier;

    modifier = (*env)->CallIntMethod( env, selfmeth, s_jcache->Method.getModifiers);
    if ((*env)->ExceptionCheck( env))
      return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), JThreadExc2py( env);
    
    isStatic = (*env)->CallStaticBooleanMethod( env,
						s_jcache->class.Modifier,
						s_jcache->Modifier.isStatic,
						modifier);
    if ((*env)->ExceptionCheck( env))
      return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), JThreadExc2py( env);
  }
  methName = (*env)->CallObjectMethod( env, selfmeth, s_jcache->Method.getName);
  methNameStr = (*env)->GetStringUTFChars( env, methName, NULL);
  {
    jobject returnType;
    returnType = (*env)->CallObjectMethod( env, selfmeth,
					   s_jcache->Method.getReturnType);
    if (NOT returnType) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
			    JThreadExc2py( env);
    
    returnName = (*env)->CallObjectMethod( env, returnType, s_jcache->Class.getName);
    if (NOT returnName) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
			    JThreadExc2py( env);
    (*env)->DeleteLocalRef( env, returnType);
  }
  returnString = (*env)->GetStringUTFChars( env, returnName, NULL);

  parameterTypes = (*env)->CallObjectMethod( env, selfmeth,
					     s_jcache->Method.getParameterTypes);
  if (NOT parameterTypes) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
			      JThreadExc2py( env);

  argc = (*env)->GetArrayLength( env, parameterTypes);

  Py_BLOCK_THREADS;
  result = PyTuple_New( 4);
  if (NOT result) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;

  if (PyTuple_SetItem( result, 0, PyInt_FromLong( isStatic)))
    return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;
  if (PyTuple_SetItem( result, 1, PyString_FromString( methNameStr)))
    return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;
  if (PyTuple_SetItem( result, 2, PyString_FromString( returnString)))
    return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;
  argtuple = PyTuple_New( argc);
  if (NOT argtuple) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;
  Py_UNBLOCK_THREADS;
  (*env)->ReleaseStringUTFChars( env, methName, methNameStr);
  (*env)->DeleteLocalRef( env, methName);
  (*env)->ReleaseStringUTFChars( env, returnName, returnString);
  (*env)->DeleteLocalRef( env, returnName);

  for (aidx = 0; aidx < argc; aidx++)
    {
      jobject arg, argstring;
      arg = (*env)->GetObjectArrayElement( env, parameterTypes, aidx);
      if (NOT arg) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
		       JThreadExc2py( env);
      
      argstring = (*env)->CallObjectMethod( env, arg, s_jcache->Class.getName);
      if (NOT argstring) return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
			     JThreadExc2py( env);
      (*env)->DeleteLocalRef( env, arg);
      
      Py_BLOCK_THREADS;
      if (PyTuple_SetItem( argtuple, aidx,
			   PyString_FromJstring( env, argstring)))
	return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), NULL;
      Py_UNBLOCK_THREADS;
      (*env)->DeleteLocalRef( env, argstring);
    }
  JPE_FRAME_CLOSE( env);
  Py_BLOCK_THREADS;
  if (PyTuple_SetItem( result, 3, argtuple)) return NULL;
  ALL_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_getClass( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jclass jres;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  JPE_FRAME_OPEN( env, 8);
    jres = (*env)->GetObjectClass( env, Py2jobject( self)->object);
    if (NOT jres) return JPE_FRAME_CLOSE( env), Jexc2py( env);
    result = jobject_new( env, jres);
    /* (*env)->DeleteLocalRef( env, jres); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_getField( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jobject field;
  string_t name;
  jstring jsname;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "s", &name)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jsname = (*env)->NewStringUTF( env, name);
      if (NOT jsname) return JPE_ExcFrame2Py( env);
      field = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					s_jcache->Class.getField, jsname);
      /* (*env)->DeleteLocalRef( env, jsname); */
      if (NOT field) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, field);
    /* (*env)->DeleteLocalRef( env, field); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_sig( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jstring jname;
  string_t sname, sig;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jname = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					s_jcache->Class.getName);
      if (NOT jname) return JPE_ExcFrame2Py( env);
      sname = (*env)->GetStringUTFChars( env, jname, NULL);
      if (NOT sname)
	{
	  /* (*env)->DeleteLocalRef( env, jname); */
	  return JPE_ExcFrame2Py( env);
	}
      sig = jutil_name2sig( env, sname);
    Py_END_ALLOW_THREADS;
    if (sig)
      {
	result = PyString_FromString( sig);
	jutil_name2sig_cleanup( &sig);
      }
    else result = Jexc2py( env);;
    (*env)->ReleaseStringUTFChars( env, jname, sname);
    /* (*env)->DeleteLocalRef( env, jname); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_superclass( PyObject_t self, PyObject_t args)
{
  jclass cls;
  JNIEnv* env;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  JPE_FRAME_OPEN( env, 8);
    cls = (*env)->GetSuperclass( env, Py2jobject( self)->object);
    if (NOT cls)
      {
	JPE_FRAME_CLOSE( env);
	return(*env)->ExceptionCheck( env)
	  ? Jexc2py( env) : (Py_INCREF( Py_None), Py_None);
      }
    result = jobject_new( env, cls);
    /* (*env)->DeleteLocalRef( env, cls); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_isAssignable( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  PyObject_t ocls2;
  jboolean res;

  if (NOT PyArg_ParseTuple( args, "O!", &PyJobject_Type, &ocls2)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}
  ALL_FRAME_OPEN( env, 8);

    res = (*env)->IsAssignableFrom( env, Py2jobject( self)->object,
				    Py2jobject( ocls2)->object);
  ALL_FRAME_CLOSE( env);
  if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
  return res ? PyInt_FromLong( 1) : PyInt_FromLong( 0);
}

static PyObject_t jobject_class_fieldID( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  string_t name, sig;
  int isStatic = 0;

  if (NOT PyArg_ParseTuple( args, "ss|i", &name, &sig, &isStatic)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  return jfieldID_new( env, isStatic, Py2jobject( self)->object, name, sig);
}
static PyObject_t jobject_class_methodID( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  string_t name, sig;
  int modifier = 0;

  if (NOT PyArg_ParseTuple( args, "ss|i", &name, &sig, &modifier)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  return jmethodID_new( env, modifier, Py2jobject( self)->object, name, sig);
}

static PyObject_t jobject_class_getMethods( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jobject methods;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      methods = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Class.getMethods);
      if (NOT methods) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, methods);
    /* (*env)->DeleteLocalRef( env, methods); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_getConstructors( PyObject_t self,
						 PyObject_t args)
{
  JNIEnv* env;
  jobject constructors;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      constructors = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					      s_jcache->Class.getConstructors);
      if (NOT constructors) return JThreadExc2py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, constructors);
    /* (*env)->DeleteLocalRef( env, constructors); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_class_methodsToStrings( PyObject_t self,
						  PyObject_t args)
{
  JNIEnv* env;
  jobjectArray methods;
  int idx, msz;
  PyObject_t result = NULL;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 32);
      methods = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Class.getMethods);
      if ((*env)->ExceptionCheck( env))
	return JPE_ExcFrame2Py( env);
      msz = (*env)->GetArrayLength( env, methods);
      if ((*env)->ExceptionCheck( env))
	return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS

      result = PyTuple_New( msz);
    
    for (idx = 0; idx < msz; idx++)
      {
	jobject meth;
	jstring mstr;
	string_t str;
	
	Py_BEGIN_ALLOW_THREADS;
	  meth = (*env)->GetObjectArrayElement( env, methods, idx);
	  mstr = (*env)->CallObjectMethod( env, meth,
					   s_jcache->Object.toString);
	  str = (*env)->GetStringUTFChars( env, mstr, NULL);
	Py_END_ALLOW_THREADS;
	PyTuple_SetItem( result, idx, PyString_FromString( str));
	(*env)->ReleaseStringUTFChars( env, mstr, str);
	(*env)->DeleteLocalRef( env, mstr);
	(*env)->DeleteLocalRef( env, meth);
	if ((*env)->ExceptionCheck( env))
	  {
	    JPE_FRAME_CLOSE( env);
	    Py_DECREF( result);
	    /* (*env)->DeleteLocalRef( env, methods); */
	    return Jexc2py( env);
	  }
      }
    /* (*env)->DeleteLocalRef( env, methods); */
  JPE_FRAME_CLOSE( env);

  return result;
}

static PyObject_t jobject_class_methodsDesc( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jobjectArray methods;
  jsize idx, msz;
  PyObject_t result = NULL;
  PyThreadState* _save;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_UNBLOCK_THREADS;
    JPE_THFRAME_OPEN( env, 32);
      methods = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Class.getMethods);
      if (NOT methods) return JPE_ExcFrame2Py( env);
      
      msz = (*env)->GetArrayLength( env, methods);
      if ((*env)->ExceptionCheck( env))
	return JPE_ExcFrame2Py( env);

    Py_BLOCK_THREADS;
    result = PyTuple_New( msz);
    Py_UNBLOCK_THREADS;

      for (idx = 0; idx < msz; idx++)
	{
	  jobject meth, parameterTypes;
	  jstring returnName, methName;
	  int isStatic;
	  jsize argc, aidx;
	  PyObject_t methdesc, argtuple;
	  string_t returnString, methNameStr;
	  
	  meth = (*env)->GetObjectArrayElement( env, methods, idx);
	  if (NOT meth) return JPE_ExcFrame2Py( env);
	  
	  {
	    jint modifier;
	    modifier = (*env)->CallIntMethod( env, meth,
					      s_jcache->Method.getModifiers);
	    if ((*env)->ExceptionCheck( env))
	      return JPE_ExcFrame2Py( env);
	    
	    isStatic = (*env)->CallStaticBooleanMethod( env,
						   s_jcache->class.Modifier,
						   s_jcache->Modifier.isStatic,
							modifier);
	    if ((*env)->ExceptionCheck( env))
	      return JPE_ExcFrame2Py( env);
	  }
	  methName = (*env)->CallObjectMethod( env, meth,
					       s_jcache->Method.getName);
	  methNameStr = (*env)->GetStringUTFChars( env, methName, NULL);
	  {
	    jobject returnType;
	    returnType = (*env)->CallObjectMethod( env, meth,
					       s_jcache->Method.getReturnType);
	    if (NOT returnType) return JPE_ExcFrame2Py( env);
	    
	    returnName = (*env)->CallObjectMethod( env, returnType,
						   s_jcache->Class.getName);
	    if (NOT returnName) return JPE_ExcFrame2Py( env);
	    (*env)->DeleteLocalRef( env, returnType);
	  }
	  returnString = (*env)->GetStringUTFChars( env, returnName, NULL);
	  
	  parameterTypes = (*env)->CallObjectMethod( env, meth,
					   s_jcache->Method.getParameterTypes);
	  if (NOT parameterTypes)
	    return JPE_ExcFrame2Py( env);
	  
	  argc = (*env)->GetArrayLength( env, parameterTypes);
	  
	Py_BLOCK_THREADS;
	methdesc = PyTuple_New( 4);
	if (NOT methdesc) return JPE_FRAME_CLOSE( env), NULL;
	
	if (PyTuple_SetItem( methdesc, 0, PyInt_FromLong( isStatic)))
	  return JPE_FRAME_CLOSE( env), NULL;
	if (PyTuple_SetItem( methdesc, 1, PyString_FromString( methNameStr)))
	  return JPE_FRAME_CLOSE( env), NULL;
	if (PyTuple_SetItem( methdesc, 2, PyString_FromString( returnString)))
	  return JPE_FRAME_CLOSE( env), NULL;
	argtuple = PyTuple_New( argc);
	if (NOT argtuple) return JPE_FRAME_CLOSE( env), NULL;
	Py_UNBLOCK_THREADS;
	  (*env)->ReleaseStringUTFChars( env, methName, methNameStr);
	  (*env)->DeleteLocalRef( env, methName);
	  (*env)->ReleaseStringUTFChars( env, returnName, returnString);
	  (*env)->DeleteLocalRef( env, returnName);

	  for (aidx = 0; aidx < argc; aidx++)
	    {
	      jobject arg, argstring;
	      string_t argstr;
	      arg = (*env)->GetObjectArrayElement( env, parameterTypes, aidx);
	      if (NOT arg) return JPE_ExcFrame2Py( env);

	      argstring = (*env)->CallObjectMethod( env, arg,
						    s_jcache->Class.getName);
	      if (NOT argstring) return JPE_ExcFrame2Py( env);
	      
	      argstr = (*env)->GetStringUTFChars( env, argstring, NULL);
	    Py_BLOCK_THREADS;
	    if (PyTuple_SetItem( argtuple, aidx, PyString_FromString( argstr)))
	      return JPE_FRAME_CLOSE( env), NULL;
	    Py_UNBLOCK_THREADS;
	      (*env)->ReleaseStringUTFChars( env, argstring, argstr);
	      (*env)->DeleteLocalRef( env, argstring);
	      (*env)->DeleteLocalRef( env, arg);
	    }
	Py_BLOCK_THREADS;
	if (PyTuple_SetItem( methdesc, 3, argtuple))
	  return JPE_FRAME_CLOSE( env), NULL;
	if (PyTuple_SetItem( result, idx, methdesc))
	  return JPE_FRAME_CLOSE( env), NULL;
	Py_UNBLOCK_THREADS;
	}
      /* (*env)->DeleteLocalRef( env, methods); */
    JPE_FRAME_CLOSE( env);
  Py_BLOCK_THREADS;

  return result;
}

static PyObject_t jobject_method_getDeclaringClass( PyObject_t self,
						    PyObject_t args)
{
  JNIEnv* env;
  jobject jresult;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jresult = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Method.getDeclaringClass);
      if (NOT jresult) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jresult);
    /* (*env)->DeleteLocalRef( env, jresult); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_method_getParameterTypes( PyObject_t self,
						    PyObject_t args)
{
  JNIEnv* env;
  jobject jresult;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jresult = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Method.getParameterTypes);
      if (NOT jresult) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jresult);
    /* (*env)->DeleteLocalRef( env, jresult); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_method_getReturnType( PyObject_t self,
						PyObject_t args)
{
  JNIEnv* env;
  jobject jresult;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jresult = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					  s_jcache->Method.getReturnType);
      if (NOT jresult) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jresult);
    /* (*env)->DeleteLocalRef( env, jresult); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_method_isStatic( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jint modifiers;
  int isStatic;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_FRAME_OPEN( env, 8);
      modifiers = (*env)->CallIntMethod( env, Py2jobject( self)->object,
					 s_jcache->Method.getModifiers);
      if ((*env)->ExceptionCheck( env))
	return JPE_FRAME_CLOSE( env), JThreadExc2py( env);
      isStatic = (*env)->CallStaticBooleanMethod( env,
						  s_jcache->class.Modifier,
						  s_jcache->Modifier.isStatic,
						  modifiers);
      if ((*env)->ExceptionCheck( env))
	return JPE_FRAME_CLOSE( env), JThreadExc2py( env);
    JPE_FRAME_CLOSE( env);
  Py_END_ALLOW_THREADS;
  return PyInt_FromLong( isStatic);
}

static PyObject_t jobject_constructor_getDeclaringClass( PyObject_t self,
							 PyObject_t args)
{
  JNIEnv* env;
  jobject jresult;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jresult = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
				      s_jcache->Constructor.getDeclaringClass);
      if (NOT jresult) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jresult);
    /* (*env)->DeleteLocalRef( env, jresult); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_constructor_getParameterTypes( PyObject_t self,
							 PyObject_t args)
{
  JNIEnv* env;
  jobject jresult;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jresult = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
				      s_jcache->Constructor.getParameterTypes);
      if (NOT jresult) return JPE_ExcFrame2Py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jresult);
    /* (*env)->DeleteLocalRef( env, jresult); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_field_isStatic( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jint modifiers;
  int isStatic;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}
  ALL_FRAME_OPEN( env, 8);

    Py_BEGIN_ALLOW_THREADS;
      modifiers = (*env)->CallIntMethod( env, Py2jobject( self)->object,
					 s_jcache->Field.getModifiers);
      if ((*env)->ExceptionCheck( env))
	return ALL_FRAME_CLOSE( env), JThreadExc2py( env);
      isStatic = (*env)->CallStaticBooleanMethod( env,
						  s_jcache->class.Modifier,
						  s_jcache->Modifier.isStatic,
						  modifiers);
    ALL_FRAME_CLOSE( env);
    if ((*env)->ExceptionCheck( env)) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  return PyInt_FromLong( isStatic);
}

static PyObject_t jobject_field_fromReflected( PyObject_t self,
					       PyObject_t args)
{
  JNIEnv* env;
  jclass jcls;
  jstring jsname;
  string_t sname, sig;
  PyObject_t result;
  jint modifier, okflag = 0;
  int isStatic;

  if (NOT PyArg_ParseTuple( args, "")) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 32);
      jcls = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
				       s_jcache->Field.getDeclaringClass);
      if (NOT jcls) {LogTrace; goto EXIT;}
      
      {
	jclass jrtype;
	jstring jsrtype;
	string_t srtype;
	
	jrtype = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					   s_jcache->Field.getType);
	if (NOT jrtype) {LogTrace; goto EXIT;}
	
	jsrtype = (*env)->CallObjectMethod( env, jrtype,
					    s_jcache->Class.getName);
	(*env)->DeleteLocalRef( env, jrtype);
	if (NOT jsrtype) {LogTrace; goto EXIT;}
	
	srtype = (*env)->GetStringUTFChars( env, jsrtype, NULL);
	if (NOT srtype) {LogTrace; goto EXIT;}
	sig = jutil_name2sig( env, srtype);
	(*env)->ReleaseStringUTFChars( env, jsrtype, srtype);
	(*env)->DeleteLocalRef( env, jsrtype);
	if (NOT sig) {LogTrace; goto EXIT;}
      }
      
      jsname = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
					 s_jcache->Field.getName);
      if (NOT jsname) {LogTrace; goto EXIT;}
      
      sname = (*env)->GetStringUTFChars( env, jsname, NULL);
      if (NOT sname) JPE_ExcFrame2Py( env);
      
      modifier = (*env)->CallIntMethod( env, Py2jobject( self)->object,
					s_jcache->Field.getModifiers);
      if (NOT modifier) {LogTrace; goto EXIT;}
      
      isStatic = (*env)->CallStaticBooleanMethod( env,
						  s_jcache->class.Modifier,
						  s_jcache->Modifier.isStatic,
						  modifier);
      if ((*env)->ExceptionCheck( env)) {LogTrace; goto EXIT;}
      
      okflag = 1;
      
 EXIT:
    Py_END_ALLOW_THREADS;
    result = okflag ?
      jfieldID_new( env, isStatic, jcls, sname, sig) : Jexc2py( env);
    jutil_name2sig_cleanup( &sig);
    (*env)->ReleaseStringUTFChars( env, jsname, sname);
    /*   (*env)->DeleteLocalRef( env, jsname); */
    /*   (*env)->DeleteLocalRef( env, jcls); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static struct PyMethodDef jobject_methods[] =
{
  {"string2UTF", jobject_string2UTF, METH_VARARGS,
   "string2UTF( self) -- jstring to UTF string -- Return PyString"},
  {"methodDesc", jobject_method_Desc, METH_VARARGS,
   "method description as tuple"},
  {"getClass", jobject_getClass, METH_VARARGS,
   "getClass( self) -- get Java class jobject"},
  {"class_getField", jobject_class_getField, METH_VARARGS,
   "class_getField( self, name) -- Class.getField()"},
  {"class_sig", jobject_class_sig, METH_VARARGS, "class signature"},
  {"GetSuperClass", jobject_class_superclass, METH_VARARGS, "get superclass"},
  {"IsAssignable", jobject_class_isAssignable, METH_VARARGS,
   "Is Assignable to Arg?"},
  {"GetFieldID", jobject_class_fieldID, METH_VARARGS, "get field ID"},
  {"getMethodID", jobject_class_methodID, METH_VARARGS, "get method ID"},
  {"getMethods", jobject_class_getMethods, METH_VARARGS,
   "Java's self.getMethods()"},
  {"getConstructors", jobject_class_getConstructors, METH_VARARGS,
   "Java's self.getConstructors()"},
  {"methodsToStrings", jobject_class_methodsToStrings, METH_VARARGS,
   "appy toStrings() to Java's self.getMethods()"},
  {"methodsDesc", jobject_class_methodsDesc, METH_VARARGS,
   "apply toStrings() to Java's self.getMethods()"},
  {"method_getDeclaringClass", jobject_method_getDeclaringClass, METH_VARARGS,
   "getDeclaringClass() on method objects"},
  {"method_getParameterTypes", jobject_method_getParameterTypes, METH_VARARGS,
   "getParameterTypes() on method objects"},
  {"method_getReturnType", jobject_method_getReturnType, METH_VARARGS,
   "getReturnType() on method objects"},
  {"method_isStatic", jobject_method_isStatic, METH_VARARGS,
   "modifier.isStatic( self.getModifiers()) on method objects"},
  {"constructor_getDeclaringClass", jobject_constructor_getDeclaringClass,
   METH_VARARGS, "getDeclaringClass() on constructor objects"},
  {"constructor_getParameterTypes", jobject_constructor_getParameterTypes,
   METH_VARARGS, "getParameterTypes() on constructor objects"},
  {"field_isStatic", jobject_field_isStatic, METH_VARARGS,
   "modifier.isStatic( self.getModifiers()) on field objects"},
  {"field_fromReflected", jobject_field_fromReflected, METH_VARARGS,
   "(*env)->FromReflectedField() on field objects"},
  {NULL, NULL}
};

static PyObject_t jobject_getattr( PyObject_t self, char* name)
{
  if (0 == strcmp( "className", name))
    {
      if (NOT Py2jobject( self)->className)
	{
	  jclass cls;
	  string_t strname;
	  JNIEnv* env;

	  env = pyJavaEnv( PyThreadState_Get());
	  if (NOT env) {LogTrace; return NULL;}

	  JPE_FRAME_OPEN( env, 8);
	    cls = (*env)->GetObjectClass( env, Py2jobject( self)->object);
	    if (NOT cls) return JPE_FRAME_CLOSE( env), Jexc2py( env);
	    strname = jutil_class_name( env, s_jcache, cls);
	    if (NOT strname) return JPE_FRAME_CLOSE( env), Jexc2py( env);
	    Py2jobject( self)->className = PyString_InternFromString( strname);
	    jutil_UTF_cleanup( &strname);
	    /* missing DeleteLocalRef call? */
	  JPE_FRAME_CLOSE( env);
	}
      return Py2jobject( self)->className;
    }
  else return Py_FindMethod( jobject_methods, self, name);
}

static int jobject_compare( PyObject_t self, PyObject_t other)
{
  JNIEnv* env;
  jboolean isSame;

  if (&PyJobject_Type != other->ob_type)
    {
      PyErr_SetString( PyExc_TypeError,
		       "jobject_compare(): other object is not PyJobject");
      return -1;
    }

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return -1;

  Py_BEGIN_ALLOW_THREADS;
    isSame = (*env)->IsSameObject( env, Py2jobject( self)->object,
				   Py2jobject( other)->object);
    if ((*env)->ExceptionCheck( env))
      {
	(void)JThreadExc2py( env);
	return -1;
      }
  Py_END_ALLOW_THREADS;
  return JNI_TRUE == isSame ? 0 : 1;
}

static PyObject_t jobject_repr( PyObject_t self)
{
  JNIEnv* env;
  jobject jstr;
  string_t str;
  PyObject_t result;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_FRAME_OPEN( env, 8);
      jstr = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
				       s_jcache->Object.toString);
      if (NOT jstr) return JPE_ExcFrame2Py( env);
      str = (*env)->GetStringUTFChars( env, jstr, NULL);
    Py_END_ALLOW_THREADS;
    result = PyString_FromString( str);
    (*env)->ReleaseStringUTFChars( env, jstr, str);
    /* (*env)->DeleteLocalRef( env, jstr); */
  JPE_FRAME_CLOSE( env);
  return result;
}

static long jobject_hash( PyObject_t self)
{
  JNIEnv* env;
  jlong result;
  
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return -1;

  Py_BEGIN_ALLOW_THREADS;
    ALL_FRAME_OPEN_TYPE( env, 8, 0);
      result = (*env)->CallLongMethod( env, Py2jobject( self)->object,
				       s_jcache->Object.hashCode);
    ALL_FRAME_CLOSE( env);
    if ((*env)->ExceptionCheck( env)) return ((void)JThreadExc2py( env), -1);
  Py_END_ALLOW_THREADS;
  return (long)result;
}

/* If the Java objet has a 'getName()' method, use it;
   else use 'toString()'
 */
static PyObject_t jobject_str( PyObject_t self)
{
  JNIEnv* env;
  jobject jstr;
  string_t str;
  PyObject_t result;
  jclass jcls;
  jmethodID getName;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_THFRAME_OPEN( env, 8);
      jcls = (*env)->GetObjectClass( env, Py2jobject( self)->object);
      if (NOT jcls) return JPE_ExcFrame2Py( env);
      getName = (*env)->GetMethodID( env, jcls, "getName",
				     "()Ljava/lang/String;");
      /* (*env)->DeleteLocalRef( env, jcls); */
      if (NOT getName)
	{
	  jthrowable exc;
	  exc = (*env)->ExceptionOccurred( env);
	  (*env)->ExceptionClear( env);
	  if (NOT (*env)->IsInstanceOf( env, exc,
					s_jcache->class.NoSuchMethodError))
	    {
	      (*env)->Throw( env, exc);
	    JThreadExc2py( env);
	  JPE_FRAME_CLOSE( env);
	  return NULL;
	    }
	  getName = s_jcache->Object.toString;
	}
      jstr = (*env)->CallObjectMethod( env, Py2jobject( self)->object,
				       getName);
      if (NOT jstr) return JPE_ExcFrame2Py( env);
      str = (*env)->GetStringUTFChars( env, jstr, NULL);
    Py_END_ALLOW_THREADS;
    result = PyString_FromString( str);
    (*env)->ReleaseStringUTFChars( env, jstr, str);
    /* (*env)->DeleteLocalRef( env, jstr); */
  JPE_FRAME_CLOSE( env);
  return result;
}

/* static PyObject_t xxxjobject_str( PyObject_t self) */
/* { */
/*   JNIEnv* env; */
/*   jobject jstr; */
/*   string_t str; */
/*   PyObject_t result; */

/*   env = pyJavaEnv( PyThreadState_Get()); */
/*   if (NOT env) {LogTrace; return NULL;} */

/*   Py_BEGIN_ALLOW_THREADS; */
/*     JPE_THFRAME_OPEN( env, 8); */
/*       jstr = (*env)->CallObjectMethod( env, Py2jobject( self)->object, */
/* 				       s_jcache->Object.toString); */
/*       if (NOT jstr) return JPE_ExcFrame2Py( env); */
/*       str = (*env)->GetStringUTFChars( env, jstr, NULL); */
/*     Py_END_ALLOW_THREADS; */
/*     result = PyString_FromString( str); */
/*     (*env)->ReleaseStringUTFChars( env, jstr, str); */
/*   JPE_FRAME_CLOSE( env); */
/*   return result; */
/* } */

/**** jobject sequence ****/

static int jobject_length( PyObject_t self)
{
  if (NOT Py2jobject( self)->type)
    { /*do required initialization*/
      JNIEnv* env;
      env = pyJavaEnv( PyThreadState_Get());
      if (NOT env) return -1;

      /*FRAME*/if ((*(env))->PushLocalFrame( env, 8)) return Jexc2py( env), -1;
        if (jobject_initType( env, Py2jobject( self)))
	  return JPE_FRAME_CLOSE( env), Jexc2py( env), -1;
      JPE_FRAME_CLOSE( env);
   }
  return  e_jarray == Py2jobject( self)->type
      ? Py2jobject( self)->arraylength : 1;

}

static PyObject_t jobject_getitem( PyObject_t self, int index)
{
  JNIEnv* env;
  jvalue value;
  
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  if (NOT Py2jobject( self)->type
      && jobject_initType( env, Py2jobject( self))) return Jexc2py( env);
  if (e_jarray != Py2jobject( self)->type)
     return PyErr_Format( PyExc_TypeError, "Not a java array");

  if (index < 0) index += Py2jobject( self)->arraylength;
  if (index < 0 || Py2jobject( self)->arraylength <= index)
    {
      PyErr_SetNone( PyExc_IndexError);
      return NULL;
    }

  switch (Py2jobject( self)->arraytype)
    {
    case e_jboolean:
      (*env)->GetBooleanArrayRegion( env, Py2jobject( self)->object, index, 1,
				     &value.z);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.z);
    case e_jbyte:
      (*env)->GetByteArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.b);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.b);
    case e_jchar:
      (*env)->GetCharArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.c);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.c);
    case e_jshort:
      (*env)->GetShortArrayRegion( env, Py2jobject( self)->object, index, 1,
				   &value.s);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.s);
    case e_jint:
      (*env)->GetIntArrayRegion( env, Py2jobject( self)->object, index, 1,
				 &value.i);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.i);
    case e_jlong:
      (*env)->GetLongArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.j);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyInt_FromLong( (long)value.j);
    case e_jfloat:
      (*env)->GetFloatArrayRegion( env, Py2jobject( self)->object, index, 1,
				   &value.f);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyFloat_FromDouble( (double)value.f);
    case e_jdouble:
      (*env)->GetDoubleArrayRegion( env, Py2jobject( self)->object, index, 1,
				    &value.d);
      if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
      return PyFloat_FromDouble( (double)value.d);
    case e_jobject:
      {
	PyObject_t result;
	
	value.l = (*env)->GetObjectArrayElement( env,
						 Py2jobject( self)->object,
						 index);
	if (value.l)
	  {
	    result = jobject_new( env, value.l);
	    (*env)->DeleteLocalRef( env, value.l);
	  }
	else
	  {
	    if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
	    LogTrace;
	    result = (Py_INCREF( Py_None), Py_None);
	  }
	return result;
      }
    default:
      return PyErr_Format( PyExc_SystemError, "incorrect array type");
    }
}

static int jobject_setitem( PyObject_t self, int index, PyObject_t obj)
{
  JNIEnv* env;
  jvalue value;
  
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return 1;

  if (NOT Py2jobject( self)->type && jobject_initType( env, Py2jobject( self)))
     SJexc2py( env)

  if (e_jarray != Py2jobject( self)->type)
      return PyErr_Format( PyExc_TypeError, "Not a java array"), 1;

  if (index < 0) index += Py2jobject( self)->arraylength;
  if (index < 0 || Py2jobject( self)->arraylength <= index)
    return PyErr_SetNone( PyExc_IndexError), 1;

  switch (Py2jobject( self)->arraytype)
    {
    case e_jboolean:
      value.z = (jboolean)(NOT NOT PyInt_AsLong( obj));
      if (PyErr_Occurred()) return 1;
      (*env)->SetBooleanArrayRegion( env, Py2jobject( self)->object, index, 1,
				     &value.z);
      break;
    case e_jbyte:
      value.b = (jbyte)PyInt_AsLong( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetByteArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.b);
      break;
    case e_jchar:
      value.c = (jchar)PyInt_AsLong( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetCharArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.c);
      break;
    case e_jshort:
      value.s = (jshort)PyInt_AsLong( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetShortArrayRegion( env, Py2jobject( self)->object, index, 1,
				   &value.s);
      break;
    case e_jint:
      value.i = (jint)PyInt_AsLong( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetIntArrayRegion( env, Py2jobject( self)->object, index, 1,
				   &value.i);
      break;
    case e_jlong:
      value.j = (jlong)PyInt_AsLong( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetLongArrayRegion( env, Py2jobject( self)->object, index, 1,
				  &value.j);
      break;
    case e_jfloat:
      value.f = (jfloat)PyFloat_AsDouble( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetFloatArrayRegion( env, Py2jobject( self)->object, index, 1,
				   &value.f);
      break;
    case e_jdouble:
      value.d = (jdouble)PyFloat_AsDouble( obj);
      if (PyErr_Occurred()) return 1;
      (*env)->SetDoubleArrayRegion( env, Py2jobject( self)->object, index, 1,
				    &value.d);
      break;
    case e_jobject:
      value.l = PyJobject_AsJobject( env, obj, NULL);
      if (NOT value.l) SJexc2py( env)
      (*env)->SetObjectArrayElement( env, Py2jobject( self)->object, index,
				     value.l);
      break;
    default:
      return PyErr_Format( PyExc_SystemError, "incorrect array type"), 1;
    }

  if ((*env)->ExceptionCheck( env)) SJexc2py( env)

  return 0;
}

static PySequenceMethods jobject_sequence = 
{
  jobject_length, /*sq_length*/
  NULL, /*sq_concat*/
  NULL, /*sq_repeat*/
  jobject_getitem, /*sq_item*/
  NULL, /*sq_slice*/
  jobject_setitem, /*sq_ass_item*/
  NULL, /*sq_ass_slice*/
};

static PyTypeObject PyJobject_Type =
{
  PyObject_HEAD_INIT( NULL)
  0, /*ob_size*/
  "jobject", /*tp_name*/
  sizeof (struct jobject), /*tb_basicsize*/
  0, /*tp_itemsize*/
  jobject_del, /*tp_dealloc*/
  NULL, /*tp_print*/
  jobject_getattr, /*tp_getattr*/
  NULL, /*tp_setattr*/
  jobject_compare, /*tp_compare*/
  jobject_repr, /*tp_repr*/
  NULL, /*tp_as_number*/
  &jobject_sequence, /*tp_as_sequence*/
  NULL, /*tp_as_mapping*/
  jobject_hash, /*tp_hash*/
  NULL, /*tp_call*/
  jobject_str, /*tp_str*/
  NULL, /*tp_getattro*/
  NULL, /*tp_setattro*/
  NULL, /*tp_as_buffer*/
  0, /*tp_xxx4*/
  "Java object", /*tp_doc*/  
};


/**** jarg ****/

static PyObject_t PyObject_FromJvalue( jvalue value, jarg_t arg)
{
  switch (arg->jtype)
    {
    case e_jboolean:
      return PyInt_FromLong( (long)value.z);
    case e_jbyte:
      return PyInt_FromLong( (long)value.b);
    case e_jchar:
      return PyInt_FromLong( (long)value.c);
    case e_jshort:
      return PyInt_FromLong( (long)value.s);
    case e_jint:
      return PyInt_FromLong( (long)value.i);
    case e_jlong:
      return PyInt_FromLong( (long)value.j);
    case e_jfloat:
      return PyFloat_FromDouble( (double)value.f);
    case e_jdouble:
      return PyFloat_FromDouble( (double)value.d);
    case e_jobject:
    case e_jarray:
      return PyErr_Format( PyExc_SystemError,
			   "PyObject_FromJvalue(): jobject and jarray types"
			   " are not covered by this function");
    default:
      return PyErr_Format( PyExc_SystemError,
			   "PyObject_FromJvalue(): uncorrect jtype");
    }
}

static jvalue PyObject_AsJvalue( JNIEnv* env, PyObject_t oval, jarg_t arg)
{
  jvalue result;

  result.i = -9999;
  switch (arg->jtype)
    {
    case e_jboolean:
      if (PyInt_Check( oval))
	result.z = (jboolean)(NOT NOT PyInt_AsLong( oval));
      else (void)PyErr_Format( PyExc_TypeError,
			       "Python IntType expected for boolean field");
      break;
    case e_jbyte:
      {
	unsigned int val;
	if (NOT PyInt_Check( oval))
	  {
	    (void)PyErr_Format( PyExc_TypeError,
				"Python IntType expected for byte field");
	    break;
	  }
	val = PyInt_AsLong( oval);
	if (val & ~0xff)
	  {
	    PyErr_SetString( PyExc_ValueError, "byte must be less than 0x100");
	    break;
	  }
	result.b = (jbyte)val;
      }
      break;
    case e_jchar:
      {
	long pval;
	jchar val;
	if (NOT PyInt_Check( oval))
	  {
	    (void)PyErr_Format( PyExc_TypeError,
				"Python IntType expected for char field");
	    break;
	  }
	pval = (jchar)PyInt_AsLong( oval);
	val = (jchar)pval;
	if (val != pval)
	  {
	    PyErr_SetString( PyExc_ValueError,
			     "Python value for java char was truncated");
	    break;
	  }
	result.c = val;
      }
      break;
    case e_jshort:
      {
	long pval;
	jshort val;
	if (NOT PyInt_Check( oval))
	  {
	    (void)PyErr_Format( PyExc_TypeError,
				"Python IntType expected for short field");
	    break;
	  }
	pval = PyInt_AsLong( oval);
	val = (jshort)pval;
	if (val != pval)
	  {
	    PyErr_SetString( PyExc_ValueError,
			     "Python value for java short was truncated");
	    break;
	  }
	result.s = val;
      }
      break;
    case e_jint:
      {
	long pval;
	jint val;
	if (NOT PyInt_Check( oval))
	  {
	    (void)PyErr_Format( PyExc_TypeError,
				"Python IntType expected for int field");
	    break;
	  }
	pval = PyInt_AsLong( oval);
	val = (jint)pval;
	if (val != pval)
	  {
	    PyErr_SetString( PyExc_ValueError,
			     "Python value for java int was truncated");
	    break;
	  }
	result.i = val;
      }
      break;
    case e_jlong:
      {
	long pval;
	jlong val;
	if (NOT PyLong_Check( oval))
	  {
	    (void)PyErr_Format( PyExc_TypeError,
				"Python LongType expected for Java long");
	    break;
	  }
	pval = PyLong_AsLong( oval);
	val = (jlong)pval;
	if (val != pval)
	  {
	    PyErr_SetString( PyExc_ValueError,
			     "Python value for Java long was truncated");
	    break;
	  }
	result.j = val;
      }
      break;
    case e_jfloat:
      if (NOT PyFloat_Check( oval))
	{
	  (void)PyErr_Format( PyExc_TypeError,
			      "Python FloatType expected for Java float");
	  break;
	}
      result.f = (jfloat)PyFloat_AsDouble( oval);
      break;
    case e_jdouble:
      if (NOT PyFloat_Check( oval))
	{
	  (void)PyErr_Format( PyExc_TypeError,
			      "Python FloatType expected for Java double");
	  break;
	}
      result.d = (jdouble)PyFloat_AsDouble( oval);
      break;
    case e_jobject:
    case e_jarray:
      if (PyJobject_Check( oval))
	{
	  Py_BEGIN_ALLOW_THREADS;
	  result.l = PyJobject_AsJobject( env, oval, arg);
	  Py_END_ALLOW_THREADS;
	  if (NOT result.l) (void)Jexc2py( env);
	}
      else if (Py_None == oval) result.l = NULL;
      else (void)PyErr_Format( PyExc_TypeError,
			       "Python JobjectType expected for Java object");
      break;
    default:
      (void)PyErr_Format( PyExc_SystemError, "%s:%i incorrect field type %c",
			  __FILE__, __LINE__, arg->jtype);
    }
  return result;
}

/**** jfieldID ****/

struct jfieldID
{
  PyObject_HEAD
  jfieldID id;
  jclass cls; /* so that class will not be unloaded while field is defined */
  int isStatic;
  struct jarg type;
};
typedef struct jfieldID* jfieldID_t;
#define Py2jfieldID( obj) ((jfieldID_t)obj)

static PyObject_t jfieldID_new( JNIEnv* env,
				int isStatic,
				jclass cls,
				string_t name,
				string_t sig)
{
  jfieldID id;
  jfieldID_t result;
  char* lsig;
  enum jtype jtype;

  Py_BEGIN_ALLOW_THREADS;
    id = (isStatic ? (*env)->GetStaticFieldID : (*env)->GetFieldID)
      ( env, cls, name, sig);
    if (NOT id) return JThreadExc2py( env);
    jtype = jutil_sig2jtype( sig);
    if (NOT jtype) return JThreadExc2py( env);
    lsig = (char*)malloc( 1 + strlen( sig));
    if (NOT lsig)
      {
	PyEval_RestoreThread( _save);
	return PyErr_Format( PyExc_MemoryError,
			     "jfieldID_new: malloc() failed");
      }
    (void)strcpy( lsig, sig);
    cls = (*env)->NewGlobalRef( env, cls);
    if (NOT cls) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;

  result = PyObject_NEW( struct jfieldID, &PyJfieldID_Type);
  if (NOT result) return NULL;
  result->isStatic = isStatic;
  result->id = id;
  result->cls = cls;
  result->type.jtype = jtype;
  result->type.sig = lsig;
  result->type.cls = NULL;

  return (PyObject_t)result;
}

static void jfieldID_del( PyObject_t self)
{
  JNIEnv* env;
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return;
  (*env)->DeleteGlobalRef( env, Py2jfieldID( self)->cls);
  free( (void*)Py2jfieldID( self)->type.sig);
  if (Py2jfieldID( self)->type.cls)
    (*env)->DeleteGlobalRef( env, Py2jfieldID( self)->type.cls);
}

static PyObject_t jfieldID_get( PyObject_t self, PyObject_t args)
{
  jvalue value;
  JNIEnv* env;
  PyObject_t jobj;

  if (NOT PyArg_ParseTuple( args, "O!",
			    &PyJobject_Type, &jobj)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    ALL_FRAME_OPEN( env, 8);
      switch (Py2jfieldID( self)->type.jtype)
	{
	case e_jboolean:
	  value.z = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticBooleanField
		     : (*env)->GetBooleanField)( env,
						 Py2jobject( jobj)->object,
						 Py2jfieldID( self)->id);
	  break;
	case e_jbyte:
	  value.b = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticByteField
		     : (*env)->GetByteField)( env, Py2jobject( jobj)->object,
					      Py2jfieldID( self)->id);
	  break;
	case e_jchar:
	  value.c = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticCharField
		     : (*env)->GetCharField)( env, Py2jobject( jobj)->object,
					      Py2jfieldID( self)->id);
	  break;
	case e_jshort:
	  value.s = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticShortField
		     : (*env)->GetShortField)( env, Py2jobject( jobj)->object,
					       Py2jfieldID( self)->id);
	  break;
	case e_jint:
	  value.i = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticIntField
		     : (*env)->GetIntField)( env, Py2jobject( jobj)->object,
					     Py2jfieldID( self)->id);
	  break;
	case e_jlong:
	  value.j = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticLongField
		     : (*env)->GetLongField)( env, Py2jobject( jobj)->object,
					      Py2jfieldID( self)->id);
	  break;
	case e_jfloat:
	  value.f = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticFloatField
		     : (*env)->GetFloatField)( env, Py2jobject( jobj)->object,
					       Py2jfieldID( self)->id);
	  break;
	case e_jdouble:
	  value.d = (Py2jfieldID( self)->isStatic
		     ? (*env)->GetStaticDoubleField
		     : (*env)->GetDoubleField)( env, Py2jobject( jobj)->object,
						Py2jfieldID( self)->id);
	  break;
	case e_jobject:
	case e_jarray:
	  JPE_THFRAME_OPEN( env, 8);
	    value.l = (Py2jfieldID( self)->isStatic
		       ? (*env)->GetStaticObjectField
		       : (*env)->GetObjectField)( env,
						  Py2jobject( jobj)->object,
						  Py2jfieldID( self)->id);
	  PyEval_RestoreThread( _save);
	  if (value.l)
	    { /* special case: delete ref and return */
	      PyObject_t result;
	      result = jobject_new( env, value.l);
	      /* (*env)->DeleteLocalRef( env, value.l); */
	    JPE_FRAME_CLOSE( env);
	  ALL_FRAME_CLOSE( env);
	  return result;
	    }
	  else
	    {
	    JPE_FRAME_CLOSE( env);
	  ALL_FRAME_CLOSE( env);
	  if ((*env)->ExceptionCheck( env)) return Jexc2py( env);
	  else return (Py_INCREF( Py_None), Py_None);
	    }
	default:
	ALL_FRAME_CLOSE( env);
      PyEval_RestoreThread( _save);
      PyErr_SetString( PyExc_SystemError, "incorrect field type");
      return NULL;
	}
    ALL_FRAME_CLOSE( env);
    if ((*env)->ExceptionCheck( env)) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  return PyObject_FromJvalue( value, &Py2jfieldID( self)->type);
}

static PyObject_t jfieldID_set( PyObject_t self, PyObject_t args)
{
  jvalue value;
  JNIEnv* env;
  PyObject_t oval, jobj = NULL;

  if (NOT PyArg_ParseTuple( args, "O!O",			    
			    &PyJobject_Type, &jobj, &oval)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  ALL_FRAME_OPEN( env, 8);
    value = PyObject_AsJvalue( env, oval, &Py2jfieldID( self)->type);
    if (PyErr_Occurred()) return ALL_FRAME_CLOSE( env), NULL;
    Py_BEGIN_ALLOW_THREADS;
      switch (Py2jfieldID( self)->type.jtype)
	{
	case e_jboolean:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticBooleanField : (*env)->SetBooleanField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.z);
	  break;
	case e_jbyte:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticByteField : (*env)->SetByteField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.b);
	  break;
	case e_jchar:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticCharField : (*env)->SetCharField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.c);
	  break;
	case e_jshort:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticShortField : (*env)->SetShortField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.s);
	  break;
	case e_jint:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticIntField : (*env)->SetIntField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.i);
	  break;
	case e_jlong:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticLongField : (*env)->SetLongField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.j);
	  break;
	case e_jfloat:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticFloatField : (*env)->SetFloatField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.f);
	  break;
	case e_jdouble:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticDoubleField : (*env)->SetDoubleField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.d);
	  break;
	case e_jobject:
	case e_jarray:
	  (Py2jfieldID( self)->isStatic
	   ? (*env)->SetStaticObjectField : (*env)->SetObjectField)
	    ( env, Py2jobject( jobj)->object, Py2jfieldID( self)->id, value.l);
	  break;
	default:
	ALL_FRAME_CLOSE( env);
      PyEval_RestoreThread( _save);
      PyErr_SetString( PyExc_SystemError, "incorrect field type");
      return NULL;
	}
    ALL_FRAME_CLOSE( env);
    if ((*env)->ExceptionCheck( env)) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  return Py_INCREF( Py_None), Py_None;
}

static struct PyMethodDef jfieldID_methods[] =
{
  {"get", jfieldID_get, METH_VARARGS, "get attribute"},
  {"set", jfieldID_set, METH_VARARGS, "set attribute"},
  {NULL, NULL}
};

static PyObject_t jfieldID_getattr( PyObject_t self, char* name)
{
  return Py_FindMethod( jfieldID_methods, self, name);
}

static int jfieldID_compare( PyObject_t self, PyObject_t other)
{
  PyErr_SetString( PyExc_NotImplementedError, "jfieldID_compare()");
  return -1;
}

static PyTypeObject PyJfieldID_Type =
{
  PyObject_HEAD_INIT( NULL)
  0, /*ob_size*/
  "jfieldID", /*tp_name*/
  sizeof (struct jfieldID), /*tb_basicsize*/
  0, /*tp_itemsize*/
  jfieldID_del, /*tp_dealloc*/
  NULL, /*tp_print*/
  jfieldID_getattr, /*tp_getattr*/
  NULL, /*tp_setattr*/
  jfieldID_compare, /*tp_compare*/
  NULL, /*tp_repr*/
  NULL, /*tp_as_number*/
  NULL, /*tp_as_sequence*/
  NULL, /*tp_as_mapping*/
  NULL, /*tp_hash*/
  NULL, /*tp_call*/
  NULL, /*tp_str*/
  NULL, /*tp_getattro*/
  NULL, /*tp_setattro*/
  NULL, /*tp_as_buffer*/
  0, /*tp_xxx4*/
  "Java Field ID", /*tp_doc*/  
};

/**** jmethodID ****/

#define IsStatic 0x1
#define IsConstructor 0x2

struct jmethodID
{
  PyObject_HEAD
  jmethodID id;
  jclass cls;
  int modifier;
  string_t sig;
  struct jarg type;
  int argc;
  jarg_t argv;
};
typedef struct jmethodID* jmethodID_t;
#define Py2jmethodID( obj) ((jmethodID_t)obj)

static PyObject_t jmethodID_new( JNIEnv* env,
				 int modifier,
				 jclass cls,
				 string_t name,
				 string_t sig)
{
  jmethodID id;
  jmethodID_t result;
  char* lsig;
  enum jtype jtype;
  string_t rsig;

  assert( NOT (modifier & IsStatic && modifier & IsConstructor));
  Py_BEGIN_ALLOW_THREADS;
    id = ((modifier & IsStatic)
	  ? (*env)->GetStaticMethodID
	  : (*env)->GetMethodID)( env, cls, name, sig);
    if (NOT id) return JThreadExc2py( env);
    lsig = (char*)malloc( 1 + strlen( sig));
    if (NOT lsig)
      {
      PyEval_RestoreThread( _save);
      return PyErr_Format( PyExc_MemoryError,
			   "jmethodID_new: malloc() failed");
      }
    (void)strcpy( lsig, sig);
    
    for (rsig = lsig; *rsig && ')' != *rsig; rsig++);
    if (NOT rsig)
      {
	free( lsig);
      PyEval_RestoreThread( _save);
      return PyErr_Format( PyExc_MemoryError,
			   "jmethodID_new: sig is not correct");
      }
    rsig++;
    
    jtype = jutil_sig2jtype( rsig);
    if (NOT jtype) { free( lsig); return JThreadExc2py( env); }
    
    cls = (*env)->NewGlobalRef( env, cls);
    if (NOT cls) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  result = PyObject_NEW( struct jmethodID, &PyJmethodID_Type);
  result->id = id;
  result->cls = cls;
  result->modifier = modifier;
  result->sig = lsig;
  result->type.jtype = jtype;
  result->type.sig = rsig;
  result->type.cls = NULL;
  result->argc = 0; /* argc and argv will be computed when needed */
  result->argv = NULL;

  return (PyObject_t)result;
}

static void jmethodID_del( PyObject_t self)
{
  JNIEnv* env;
  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return;

  Py_BEGIN_ALLOW_THREADS;
    (*env)->DeleteGlobalRef( env, Py2jmethodID( self)->cls);
    free( (void*)Py2jmethodID( self)->sig);
    /* rsig points on the end of sig */
    if (Py2jmethodID( self)->argv)
      {
	jarg_t arg;
	for (arg = Py2jmethodID( self)->argv; arg->sig; arg++)
	  {
	    free( (void*)arg->sig);
	    if (arg->cls) (*env)->DeleteGlobalRef( env, arg->cls);
	  }
	free( Py2jmethodID( self)->argv);
      }
  Py_END_ALLOW_THREADS;
}

static PyObject_t jmethodID_call( PyObject_t self,
				  PyObject_t args, PyObject_t kw)
{
  JNIEnv* env;
  PyObject_t jobj, oargs = NULL;
  jarg_t argv;
  jvalue* vargs;
  jvalue retval;

  if (NOT PyArg_ParseTuple( args, "O!|O!",
			    &PyJobject_Type, &jobj,
			    &PyTuple_Type, &oargs)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  ALL_FRAME_OPEN( env, 8);
    argv = Py2jmethodID( self)->argv;
    if (NOT argv)
      { /* compute arguments definition */
	string_t sigstart, sigend;
	int idx;
	Py_BEGIN_ALLOW_THREADS;
	  sigstart = Py2jmethodID( self)->sig;
	  if ('(' != *sigstart)
	    {
	    ALL_FRAME_CLOSE( env);
	  PyEval_RestoreThread( _save);
	  return  PyErr_Format( PyExc_SystemError, "sig must start with '('");
	    }
	  sigstart++;
	  for (idx = 0; *sigstart && ')' != *sigstart; idx++)
	    {
	      jarg_t arg;
	      switch (*sigstart)
		{
		case 'Z':
		case 'B':
		case 'C':
		case 'S':
		case 'I':
		case 'J':
		case 'F':
		case 'D':
		  sigend = sigstart + 1;
		  break;
		case 'L':
		case '[':
		  for (sigend = sigstart+1; *sigend && ';' != *sigend;
		       sigend++);
		  if (';' != *sigend)
		    {
		    ALL_FRAME_CLOSE( env);
		  PyEval_RestoreThread( _save);
		  return PyErr_Format( PyExc_SystemError,
				    "jmethodID_call(): incorrect signature I");
		    }
		  sigend++;
		  break;
		default:
		ALL_FRAME_CLOSE( env);
	      PyEval_RestoreThread( _save);
	      return PyErr_Format( PyExc_SystemError,
				   "jmethodID_call(): incorrect signature");
		}
	      if (argv) argv = (jarg_t)realloc( argv, (idx+2)*sizeof *argv);
	      else argv = (jarg_t)calloc( idx + 2, sizeof *argv);
	      arg = argv + idx;
	      argv[ idx + 1] = Jarg_null;
	      {
		int slen;
		char* sig;
		slen = sigend - sigstart;
		arg->jtype = jutil_sig2jtype( sigstart);
		if ('L' == *sigstart)
		  { /* transform 'Lxxxx;' to 'xxxx' */
		    assert( ';' == *(sigend - 1));
		    sigstart++;
		    slen -= 2;
		  }
		sig = (char*)malloc( slen + 1);
		(void)strncpy( sig, sigstart, slen);
		sig[ slen] = '\0';
		arg->sig = sig;
		arg->cls = NULL;
	      }
	      sigstart = sigend;
	    }
	  Py2jmethodID( self)->argc = idx;
	  if (NOT argv)
	    {
	      assert(0 == Py2jmethodID( self)->argc);
	      argv = (jarg_t)calloc( 1, sizeof *argv);
	      *argv = Jarg_null;
	    }
	  Py2jmethodID( self)->argv = argv;
	Py_END_ALLOW_THREADS;
      }

    if (NOT oargs || NOT PyTuple_Size( oargs)) vargs = NULL;
    else
      {
	int argc, idx;
	
	argc = PyTuple_Size( oargs);
	if (argc != Py2jmethodID( self)->argc)
	  return ALL_FRAME_CLOSE( env),
	    PyErr_Format( PyExc_TypeError, "incorrect number of arguments"
			  "%i expected, %i given",
			  Py2jmethodID( self)->argc, argc);

	vargs = (jvalue*)calloc( 1 + argc, sizeof *vargs);
	for (idx = 0; idx < argc; idx++)
	  {
	    PyObject_t oval;
	    
	    oval = PyTuple_GetItem( oargs, idx);
	    if (NOT oval) {free( vargs);  return ALL_FRAME_CLOSE( env), NULL;}
	    vargs[ idx] = PyObject_AsJvalue( env, oval, &argv[ idx]);
	    if (PyErr_Occurred())
	      {
		free( vargs);
	      ALL_FRAME_CLOSE( env);
	      return NULL;
	      }
	  }
      }

    retval.i = -9999;
    Py_BEGIN_ALLOW_THREADS;
      switch (Py2jmethodID( self)->type.jtype)
	{
	case e_void:
	  {
	    int modifier;
	    modifier = Py2jmethodID( self)->modifier;
	    if (modifier & IsConstructor)
	      {
		PyObject_t result;
		JPE_THFRAME_OPEN( env, 8);
		  retval.l = (*env)->NewObjectA( env,
						 Py2jobject( jobj)->object,
						 Py2jmethodID( self)->id,
						 vargs);
		  if (vargs) free( vargs);
		  if ((*env)->ExceptionCheck( env))
		    return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env),
		      JThreadExc2py( env);
		PyEval_RestoreThread( _save);
		result = jobject_new( env, retval.l);
		/* (*env)->DeleteLocalRef( env, retval.l); */
	      JPE_FRAME_CLOSE( env);
	    ALL_FRAME_CLOSE( env);
	    return result;
	      }
	    else
	      {
		((modifier & IsStatic)
		 ? (*env)->CallStaticVoidMethodA : (*env)->CallVoidMethodA)
		  ( env, Py2jobject( jobj)->object,
		    Py2jmethodID( self)->id, vargs);
		if (vargs) free( vargs);
		if ((*env)->ExceptionCheck( env))
		  return ALL_FRAME_CLOSE( env), JThreadExc2py( env);
	      PyEval_RestoreThread( _save);
	      return ALL_FRAME_CLOSE( env), (Py_INCREF( Py_None), Py_None);
	      }
	  }
	case 'L':
	case '[':
	  JPE_THFRAME_OPEN( env, 8);
	    retval.l = ((Py2jmethodID( self)->modifier & IsStatic)
			? (*env)->CallStaticObjectMethodA
			: (*env)->CallObjectMethodA)
	      ( env, Py2jobject( jobj)->object,
		Py2jmethodID( self)->id, vargs);
	    if (vargs) free( vargs);
	  PyEval_RestoreThread( _save);
	  if ((*env)->ExceptionCheck( env))
	    return JPE_FRAME_CLOSE( env), ALL_FRAME_CLOSE( env), Jexc2py( env);
	  else if (retval.l)
	    {
	      PyObject_t result;
	      result = jobject_new( env, retval.l);
	      /* (*env)->DeleteLocalRef( env, retval.l); */
	    JPE_FRAME_CLOSE( env);
	  ALL_FRAME_CLOSE( env);
	  return result;
	    }
	  else
	    {
	    JPE_FRAME_CLOSE( env);
	  ALL_FRAME_CLOSE( env);
	  return (Py_INCREF( Py_None), Py_None);
	    }
	case e_jboolean:
	  retval.z = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticBooleanMethodA
		      : (*env)->CallBooleanMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jbyte:
	  retval.b = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticByteMethodA
		      : (*env)->CallByteMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jchar:
	  retval.c = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticCharMethodA
		      : (*env)->CallCharMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jshort:
	  retval.s = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticShortMethodA
		      : (*env)->CallShortMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jint:
	  retval.i = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticIntMethodA
		      : (*env)->CallIntMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jlong:
	  retval.j = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticLongMethodA
		      : (*env)->CallLongMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jfloat:
	  retval.f = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticFloatMethodA
		      : (*env)->CallFloatMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	case e_jdouble:
	  retval.d = (Py2jmethodID( self)->modifier & IsStatic
		      ? (*env)->CallStaticDoubleMethodA
		      : (*env)->CallDoubleMethodA)
	    ( env, Py2jobject( jobj)->object, Py2jmethodID( self)->id, vargs);
	  break;
	default:
	ALL_FRAME_CLOSE( env);
	if (vargs) free( vargs);
      PyEval_RestoreThread( _save);
      PyErr_SetString( PyExc_SystemError, "incorrect field signature");
      return NULL;
	}
    ALL_FRAME_CLOSE( env);
    if (vargs) free( vargs);
    if ((*env)->ExceptionCheck( env)) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  return PyObject_FromJvalue( retval, &Py2jmethodID( self)->type);
}

static struct PyMethodDef jmethodID_methods[] =
{
  {NULL, NULL}
};

static PyObject_t jmethodID_getattr( PyObject_t self, char* name)
{
  return Py_FindMethod( jmethodID_methods, self, name);
}

static PyTypeObject PyJmethodID_Type =
{
  PyObject_HEAD_INIT( NULL)
  0, /*ob_size*/
  "jmethodID", /*tp_name*/
  sizeof (struct jmethodID), /*tb_basicsize*/
  0, /*tp_itemsize*/
  jmethodID_del, /*tp_dealloc*/
  NULL, /*tp_print*/
  jmethodID_getattr, /*tp_getattr*/
  NULL, /*tp_setattr*/
  NULL, /*tp_compare*/
  NULL, /*tp_repr*/
  NULL, /*tp_as_number*/
  NULL, /*tp_as_sequence*/
  NULL, /*tp_as_mapping*/
  NULL, /*tp_hash*/
  jmethodID_call, /*tp_call*/
  NULL, /*tp_str*/
  NULL, /*tp_getattro*/
  NULL, /*tp_setattro*/
  NULL, /*tp_as_buffer*/
  0, /*tp_xxx4*/
  "Java Method ID", /*tp_doc*/  
};

static PyObject_t jobject_jclass( PyObject_t self, PyObject_t args)
{
  string_t name;
  jclass cls;
  JNIEnv* env;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "s", &name)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}
  Py_BEGIN_ALLOW_THREADS;
    cls = (*env)->FindClass( env, name);
    if (NOT cls) return JThreadExc2py( env);
  Py_END_ALLOW_THREADS;
  result = jobject_new( env, cls);
  (*env)->DeleteLocalRef( env, cls);
  return result;
}

static PyObject_t jclass_forName( PyObject_t self, PyObject_t args)
{
  string_t name;
  jclass cls;
  JNIEnv* env;
  jstring jname;
  int initializer = 0;
  PyObject_t oloader = NULL, result;

  if (NOT PyArg_ParseTuple( args, "s|iO!", &name, &initializer,
			    &PyJobject_Type, &oloader)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_FRAME_OPEN( env, 8);
      jname = (*env)->NewStringUTF( env, name);
      if (oloader)
	{
	  cls = (*env)->CallStaticObjectMethod( env,
						s_jcache->class.Class,
						s_jcache->Class.forNameLoader,
						jname,
						(jboolean)initializer,
						Py2jobject( oloader)->object);
	}
      else
	{
	  cls = (*env)->CallStaticObjectMethod( env,
						s_jcache->class.Class,
						s_jcache->Class.forName,
						jname);
	}
      if (NOT cls) return JPE_FRAME_CLOSE( env), JThreadExc2py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, cls);
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t javalib_jstring( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jstring jres;
  string_t str;
  PyObject_t result;

  if (NOT PyArg_ParseTuple( args, "s", &str)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  JPE_FRAME_OPEN( env, 8);
    jres = (*env)->NewStringUTF( env, str);
    if (NOT jres) return JPE_FRAME_CLOSE( env), Jexc2py( env);
    result = jobject_new( env, jres);
  JPE_FRAME_CLOSE( env);
  return result;
}

static int jclinit( JNIEnv* env, jclass* pcl, jmethodID* pclinit,
		    string_t clname, string_t initsig)
{ /* called from jobject_fromObject; use only Java context */
  assert(NOT *pclinit);
  if (NOT *pcl)
    {
      *pcl = jutil_global_class( env, clname);
      if (NOT *pcl) {LogTrace; return 1;}
    }

  *pclinit = (*env)->GetMethodID( env, *pcl, "<init>", initsig);
  if (NOT *pclinit)
    {
      LogTrace;
#ifdef JPE_DEBUG
      (void)fprintf( stderr, "%s:%i ERROR: GetMethodID( '%s', '<init', '%s') "
		     "failed\n",
		     __FILE__, __LINE__, clname, initsig);
#endif
      return 1;
    }
  return 0;
}

jobject jobject_fromObject( JNIEnv* env, PyObject_t obj)
{
  jobject result;
  PyTypeObject* typ;

  typ = obj->ob_type;

#define JPEMakeNew( name)						\
  {									\
    static jclass cl = NULL;						\
    static jmethodID clinit = NULL;					\
    if (NOT clinit && jclinit( env, &cl, &clinit, name, "(J)V"))	\
      {LogTrace; return NULL;}						\
    result = (*env)->NewObject( env, cl, clinit, (jlong)obj);		\
  }

  if (typ == &PyString_Type) JPEMakeNew( "python/PyString")
  else if (typ == &PyInstance_Type) JPEMakeNew( "python/PyInstance")
  else if (typ == &PyDict_Type) JPEMakeNew( "python/PyMapping")
  else if (typ == &PyModule_Type) JPEMakeNew( "python/PyModule")
  else if (typ == &PyList_Type) JPEMakeNew( "python/PyList")
  else if (typ == &PyTuple_Type) JPEMakeNew( "python/PyTuple")
  else if (typ == &PyJobject_Type)
    {result = PyJobject_AsJobject( env, obj, NULL);}
  else JPEMakeNew( "python/PyObject")
#undef MakeNew

  return result;
}

static PyObject_t javalib_jobjFromObject( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  jobject jres;
  PyObject_t obj, result;

  if (NOT PyArg_ParseTuple( args, "O", &obj)) return NULL;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  Py_BEGIN_ALLOW_THREADS;
    JPE_FRAME_OPEN( env, 8);
      jres = jobject_fromObject( env, obj);
      if (NOT jres) return JPE_FRAME_CLOSE( env), JThreadExc2py( env);
    Py_END_ALLOW_THREADS;
    result = jobject_new( env, jres);
  JPE_FRAME_CLOSE( env);
  return result;
}

static PyObject_t jobject_makeClass( PyObject_t self, PyObject_t args)
{
  JNIEnv* env;
  PyObject_t pyobj;
  jobject jobj;

  static struct
  {
    int initialized;
    PyObject_t object, string, array, list, map, stream;
  } makeClass = {0, NULL, NULL, NULL, NULL, NULL, NULL};

  if (NOT PyArg_ParseTuple( args, "O!", &PyJobject_Type, &pyobj)) return NULL;
  jobj = Py2jobject( pyobj)->object;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) {LogTrace; return NULL;}

  if (NOT makeClass.initialized)
    {
      PyObject_t jomod = NULL;

      jutil_rlock_take( env, s_jrlock);
      jomod = PyImport_ImportModule( "java.jobject");
      if (NOT jomod) {LogTrace; goto EXIT;}

      makeClass.object = PyObject_GetAttrString( jomod, "Jobject");
      if (NOT makeClass.object) {LogTrace; goto EXIT;}

      makeClass.string = PyObject_GetAttrString( jomod, "Jstring");
      if (NOT makeClass.string) {LogTrace; goto EXIT;}

      makeClass.array = PyObject_GetAttrString( jomod, "Jarray");
      if (NOT makeClass.array) {LogTrace; goto EXIT;}

      makeClass.list = PyObject_GetAttrString( jomod, "Jlist");
      if (NOT makeClass.list) {LogTrace; goto EXIT;}

      makeClass.map = PyObject_GetAttrString( jomod, "Jmap");
      if (NOT makeClass.map) {LogTrace; goto EXIT;}

      makeClass.stream = PyObject_GetAttrString( jomod, "Jstream");
      if (NOT makeClass.stream) {LogTrace; goto EXIT;}

      makeClass.initialized = 1;

    EXIT:
      if (NOT makeClass.initialized)
	{
	  LogTrace;
	  Py_XDECREF( makeClass.object);
	  Py_XDECREF( makeClass.string);
	  Py_XDECREF( makeClass.array);
	  Py_XDECREF( makeClass.list);
	  Py_XDECREF( makeClass.map);
	  Py_XDECREF( makeClass.stream);
	}
      jutil_rlock_release( env, s_jrlock);
      Py_XDECREF( jomod);
    }

  Py_BEGIN_ALLOW_THREADS;
    if ((*env)->IsInstanceOf( env, jobj, s_jcache->class.String))
      {
      PyEval_RestoreThread(_save);
      return Py_INCREF( makeClass.string), makeClass.string;
      }
    if ((*env)->IsInstanceOf( env, jobj, s_jcache->class.Map))
      {
      PyEval_RestoreThread(_save);
      return Py_INCREF( makeClass.map), makeClass.map;
      }
    if ((*env)->IsInstanceOf( env, jobj, s_jcache->class.List))
      {
      PyEval_RestoreThread(_save);
      return Py_INCREF( makeClass.list), makeClass.list;
      }
    if ((*env)->IsInstanceOf( env, jobj, s_jcache->class.InputStream)
	|| (*env)->IsInstanceOf( env, jobj, s_jcache->class.OutputStream))
      {
      PyEval_RestoreThread(_save);
      return Py_INCREF( makeClass.stream), makeClass.stream;
      }
  Py_END_ALLOW_THREADS;

  if (NOT Py2jobject( pyobj)->type
      && jobject_initType( env, Py2jobject( pyobj)))
      return Jexc2py( env);
  if (e_jarray == Py2jobject( pyobj)->type)
    {
      return Py_INCREF( makeClass.array), makeClass.array;
    }
  assert( e_jobject == Py2jobject( pyobj)->type);
  return Py_INCREF( makeClass.object), makeClass.object;
}

struct PyMethodDef jvmmodule_methods[] =
{
  {"isInitialized", jvm_isInitialized, METH_VARARGS, "isInitalized()\n\
  return (boolean) is a jvm present?"},
  {"create", jvm_init, METH_VARARGS, "create( module, libname, opts)\n\
   start a Java VM object;\n\
   return tuple (JVM, JENV)"},
  {"jclass", jobject_jclass, METH_VARARGS,
   "find jclass [relies on (*env)->findClass()]"},
  {"jclassForName", jclass_forName, METH_VARARGS,
   "jclass object from java.lang.class.forName()"},
  {"jstring", javalib_jstring, METH_VARARGS, "create jstring"},
  {"jproxy", javalib_jobjFromObject, METH_VARARGS,
   "jproxy( obj): create a Java object (jobject from a Python object)\n\
  Arguments:\n\
    obj -- Object : arbitrary Python object\n\
  Return:\n\
    -- JobjectType: proxy to Python object, or java object itself"},
  {"makeClass", jobject_makeClass, METH_VARARGS,
   "makeClass( jobject)\n\
    return String key for java.jobject.jobjMap dict\n\
  Argument: jobject -- JobjectType: low-level jobject wrapper"},
  {NULL, NULL}
};

void jpe_initTypes( void)
{
  PyJvm_Type.ob_type
    = PyJfieldID_Type.ob_type
    = PyJmethodID_Type.ob_type
    = PyJobject_Type.ob_type
    = &PyType_Type;
}

/* set javalib type constants */
int javalib_initJvm( PyObject_t module)
{
  JNIEnv* env;

  if (PyObject_SetAttrString( module,
			      "JobjectType", (PyObject_t)&PyJobject_Type))
    return 1;

  if (NOT s_JVM) return 0;

  env = pyJavaEnv( PyThreadState_Get());
  if (NOT env) return 1;

  return javalib_env_initJvm( env, module);
}

PyObject_t PyJinstance_FromJobject( JNIEnv* env, jobject jobj)
{
  static PyObject_t s_makeJobject = NULL;

  PyObject_t pobj, result;

  if (NOT s_makeJobject)
    {
      PyObject_t javaMod;
      javaMod = PyImport_ImportModule( "java");
      if (NOT javaMod) return NULL;
      s_makeJobject = PyObject_GetAttrString( javaMod, "MakeJobject");
      Py_DECREF( javaMod);
      if (NOT s_makeJobject) return NULL;
    }

  pobj = jobject_new( env, jobj);
  if (NOT pobj) return NULL;
  result = PyObject_CallFunction( s_makeJobject, "O", pobj);
  Py_DECREF( pobj);
  return result;
}
