/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is "Java-Python Extension (JPE)".
 * 
 * The Initial Developer of the Original Code is Frederic Bruno Giacometti.
 * Portions created by Frederic Bruno Giacometti are
 * Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
 * 
 * Contributor(s): frederic.giacometti@arakne.com
 *
 * Acknowledgments:
 * Particular gratitude is expressed to the following parties for their
 * contributing support to the development of JPE:
 *     - Arakne SARL (France) and its staff (http://www.arakne.com);
 *       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
 *     - The European Union through its funding of the IST Alice project
 *     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
 *       coordinator of the IST Alice project
 *     - iPhrase Technologies Inc. (Massachusetts, USA)
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <jni.h>
#include "jutil.h"

void jutil_cache_cleanup( struct jcache** pcache)
{
  /* +++ release fields */

  free( *pcache);
  *pcache = NULL;
}

struct jcache* jutil_cache_create( JNIEnv* env)
{
  struct jcache* result;

  result = (struct jcache*)calloc( sizeof *result, 1);
  assert( NOT result->initialized);

#define InClass( field, name)					\
  result->class.field = jutil_global_class( env, (name));	\
  if (NOT result->class.field) {LogTrace; goto EXIT;};

#define InMethod( cname, mname, sig)					\
  result->cname.mname = (*env)->GetMethodID( env, result->class.cname,	\
					     #mname, (sig));		\
  if (NOT result->cname.mname) {LogTrace; goto EXIT;};

#define InStaticMethod( cname, mname, sig)				     \
  result->cname.mname = (*env)->GetStaticMethodID( env, result->class.cname, \
					     #mname, (sig));		     \
  if (NOT result->cname.mname) {LogTrace; goto EXIT;};

#define InField( cname, fname, typ)				    \
  result->cname.fname = (*env)->GetFieldID( env, result->class.cname, \
					          #fname, (typ));	    \
  if (NOT result->cname.fname) {LogTrace; goto EXIT;};

#define InStaticField( cname, fname, typ)				    \
  result->cname.fname = (*env)->GetStaticFieldID( env, result->class.cname, \
					          #fname, (typ));	    \
  if (NOT result->cname.fname) {LogTrace; goto EXIT;};

  InClass( Throwable, "java/lang/Throwable");
  InMethod( Throwable, printStackTrace, "()V");

  InClass( PyThreadState,  "python/PyThreadState");
  InStaticField( PyThreadState, pyInterp, "J");
  InStaticMethod( PyThreadState, attach, "(JZ)V");

  InClass( Object, "java/lang/Object");
  InMethod( Object, hashCode, "()I");
  InMethod( Object, toString, "()Ljava/lang/String;");

  InClass( Class, "java/lang/Class");
  InStaticMethod( Class, forName, "(Ljava/lang/String;)Ljava/lang/Class;");
  result->Class.forNameLoader = (*env)->GetStaticMethodID( env,
							   result->class.Class,
							   "forName",
	     "(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;");
  if (NOT result->Class.forNameLoader) {LogTrace; goto EXIT;}
  InMethod( Class, getClassLoader, "()Ljava/lang/ClassLoader;");
  InMethod( Class, getConstructors, "()[Ljava/lang/reflect/Constructor;");
  InMethod( Class, getField, "(Ljava/lang/String;)Ljava/lang/reflect/Field;");
  InMethod( Class, getMethods, "()[Ljava/lang/reflect/Method;");
  InMethod( Class, getName, "()Ljava/lang/String;");
  InMethod( Class, isArray, "()Z");

#define PrimType( lname, sname)						\
  {									\
    jclass classtype;						        \
    jfieldID typeField;							\
									\
    classtype = jutil_local_class( env, "java/lang/" #lname);		\
    typeField = (*env)->GetStaticFieldID( env, classtype, "TYPE",	\
					  "Ljava/lang/Class;");		\
    if (NOT typeField) {LogTrace; goto EXIT;}				\
									\
    result->class.sname = (*env)->GetStaticObjectField( env, classtype, \
							typeField);	\
    (*env)->DeleteLocalRef( env, classtype);				\
    if (NOT result->class.sname) {LogTrace; goto EXIT;}			\
    jutil_make_global( env, &result->class.sname);		        \
  }

  PrimType( Void, V);
  PrimType( Boolean, Z);
  PrimType( Byte, B);
  PrimType( Character, C);
  PrimType( Short, S);
  PrimType( Integer, I);
  PrimType( Long, J);
  PrimType( Float, F);
  PrimType( Double, D);

  InClass( L, "java/lang/Object");
  InClass( String, "java/lang/String");
  InClass( List, "java/util/List");
  InClass( Map, "java/util/Map");
  InClass( InputStream, "java/io/InputStream");
  InClass( OutputStream, "java/io/OutputStream");
  InClass( ClassCastException, "java/lang/ClassCastException");
  InClass( NoSuchMethodError, "java/lang/NoSuchMethodError");

  InClass( Field, "java/lang/reflect/Field");
  InMethod( Field, getDeclaringClass, "()Ljava/lang/Class;");
  InMethod( Field, getModifiers, "()I");
  InMethod( Field, getName, "()Ljava/lang/String;");
  InMethod( Field, getType, "()Ljava/lang/Class;");

  InClass( Method, "java/lang/reflect/Method");
  InMethod( Method, getDeclaringClass, "()Ljava/lang/Class;");
  InMethod( Method, getModifiers, "()I");
  InMethod( Method, getName, "()Ljava/lang/String;");
  InMethod( Method, getParameterTypes, "()[Ljava/lang/Class;");
  InMethod( Method, getReturnType, "()Ljava/lang/Class;");

  InClass( Constructor, "java/lang/reflect/Constructor");
  InMethod( Constructor, getDeclaringClass, "()Ljava/lang/Class;");
  InMethod( Constructor, getModifiers, "()I");
  InMethod( Constructor, getName, "()Ljava/lang/String;");
  InMethod( Constructor, getParameterTypes, "()[Ljava/lang/Class;");

  InClass( Modifier, "java/lang/reflect/Modifier");
  InStaticMethod( Modifier, isStatic, "(I)Z");

  result->initialized = 1;

 EXIT:
  if (NOT result->initialized) jutil_cache_cleanup( &result);
  return result;
}

void jutil_rlock_take( JNIEnv* env, jobject jrlock)
{
  assert( jrlock);
#ifdef JUTIL_DEBUG
  (void)fprintf( stderr, ">>> %s:%i take rlock 0x%x\n",
		 __FILE__, __LINE__, jrlock);
#endif
  if ((*env)->MonitorEnter( env, jrlock))
    {
      (void)fprintf( stderr, "%s:%i failed to take rlock\n",
		     __FILE__, __LINE__);
      if ((*env)->ExceptionCheck( env)) (*env)->ExceptionDescribe( env);
      else (*env)->FatalError( env, __FILE__ "MonitorEnter failed");
    }
}

void jutil_rlock_release( JNIEnv* env, jobject jrlock)
{
  assert( jrlock);
#ifdef JUTIL_DEBUG
  (void)fprintf( stderr, "<<< %s:%i release rlock 0x%x\n",
		 __FILE__, __LINE__, jrlock);
#endif
  if ((*env)->MonitorExit( env, jrlock))
    {
      (void)fprintf( stderr, "%s:%i failed to release rlock\n",
		     __FILE__, __LINE__);
      if ((*env)->ExceptionCheck( env)) (*env)->ExceptionDescribe( env);
      else (*env)->FatalError( env, __FILE__ "MonitorExit failed");
    }
}

jclass jutil_local_class( JNIEnv* env, char const* name)
{
  jclass result;
  
  result = (*env)->FindClass( env, name);
  if (NOT result)
    {
      (void)fprintf( stderr, "%s:%i jutil_local_class(): "
		     "FindClass( 0x%x, 0x%x) failed on class <%s>\n",
		     __FILE__, __LINE__, env, name, name ? name : "NULL");
      (*env)->ExceptionDescribe( env);
      (*env)->FatalError( env, __FILE__);
    }
  return result;
}

jclass jutil_global_class( JNIEnv* env, char const* name)
{
  jclass result;
  
  result = (*env)->FindClass( env, name);
  if (NOT result) return NULL;
  jutil_make_global( env, &result);
  return result;
}

char const* jutil_UTF_from_jstring( JNIEnv* env, jstring jstr)
{
  char* result;
  jsize slen;

  slen = (*env)->GetStringUTFLength( env, jstr);
  if (slen < 0 || (*env)->ExceptionCheck( env)) return NULL;
  result = (char*)malloc( 1 + slen);
  if (NOT result)
    {
      LogTrace;
      (*env)->FatalError( env, "jstring_AsString(): malloc() failed");
    }
  (*env)->GetStringUTFRegion( env, jstr, 0, slen, result);
  return result;
}

void jutil_UTF_cleanup( char const** pstr)
{
  if (*pstr)
    {
      free( (void*)*pstr);
      *pstr = NULL;
    }
}

char const* jutil_class_name( JNIEnv* env, struct jcache* jcache, jclass cls)
{
  jstring jres;
  char const* result;

  jres = (*env)->CallObjectMethod( env, cls, jcache->Class.getName);
  if (NOT jres) return NULL;

  result = jutil_UTF_from_jstring( env, jres);
  (*env)->DeleteLocalRef( env, jres);

  return result;
}

void jutil_make_global( JNIEnv* env, jobject* pref)
{
  jobject globalref;
  assert( *pref);
  globalref = (*env)->NewGlobalRef( env, *pref);
  (*env)->DeleteLocalRef( env, *pref);
  if (NOT globalref)
    {
      LogTrace;
      (*env)->FatalError( env, __FILE__ " jutil_make_global(): "
			  "NewGlobalRef() failed");
    }
  *pref = globalref;
}

void jutil_throw( JNIEnv* env, jclass err, char const* msg)
{
  if ((*env)->ThrowNew( env, err, msg))
    {
      (void)fprintf( stderr, "%s:%i ERROR jutil_throw(): "
		     "ThrowNew( 0x%x, 0x%x, 0x%s) failed\n",
		     __FILE__, __LINE__, env, err, msg);
      if ((*env)->ExceptionCheck( env))
	{(*env)->ExceptionDescribe( env);}
      else
	{(*env)->FatalError( env,
			     "ThrowNew( env, s_jclass->class."
			     "ClassCastException, msg) failed");}
    }
}

void jutil_throw_illegal_state( JNIEnv* env, char const* msg)
{
  jclass err;
  err = jutil_local_class( env, "java/lang/IllegalStateException");
  jutil_throw( env, err, msg);
  (*env)->DeleteLocalRef( env, err);
}

char const* jutil_name2sig( JNIEnv* env, char const* name)
{
  enum jtype jtype;

  if (e_jarray == *name) jtype = e_jarray;
  else
    {
      char const* ptr;
      for (ptr = name; *ptr && '.' != *ptr; ptr++);
      if (*ptr) jtype = e_jobject;
      else if (NOT strcmp( "void", name)) jtype = e_void;
      else if (NOT strcmp( "boolean", name)) jtype = e_jboolean;
      else if (NOT strcmp( "byte", name)) jtype = e_jbyte;
      else if (NOT strcmp( "char", name)) jtype = e_jchar;
      else if (NOT strcmp( "short", name)) jtype = e_jshort;
      else if (NOT strcmp( "int", name)) jtype = e_jint;
      else if (NOT strcmp( "long", name)) jtype = e_jlong;
      else if (NOT strcmp( "float", name)) jtype = e_jfloat;
      else if (NOT strcmp( "double", name)) jtype = e_jdouble;
      else
	{
	  char msg[ 1024];
	  LogTrace;
	  (void)sprintf( msg, "jutil_name2jtype(): incorrect type name <%s>",
			name);
	  jutil_throw_illegal_state( env, msg);
	  return NULL;
	}
    }

  switch (jtype)
    {
    case e_void:
      return "V";
    case e_jboolean:
      return "Z";
    case e_jbyte:
      return "B";
    case e_jchar:
      return "C";
    case e_jshort:
      return "S";
    case e_jint:
      return "I";
    case e_jlong:
      return "J";
    case e_jfloat:
      return "F";
    case e_jdouble:
      return "D";
    case e_jobject:
      {
	char* result, *ptr;
	result = (char*)malloc( 3 + strlen( name));
	*result = 'L';
	(void)strcpy( result + 1, name);
	(void)strcat( result, ";");
	for (ptr = result; *ptr; ptr++)
	  if ('.' == *ptr) *ptr = '/';
	assert( ';' == result[ strlen( result) - 1]);
	return result;
      }
    case e_jarray:
      {
	char* result, *ptr;
	assert( '[' == *name);
	result = (char*)malloc( 1 + strlen( name));
	(void)strcpy( result, name);
	for (ptr = result; *ptr; ptr++)
	  if ('.' == *ptr) *ptr = '/';
	return result;
      }
    default:
      {
	LogTrace;
	jutil_throw_illegal_state( env, __FILE__ 
				   " jutil_name2sig: undefined jtype");
	return NULL;
      }
    }
}

void jutil_name2sig_cleanup( char const** psig)
{
  if (NOT *psig)
    {
      (void)fprintf( stderr, "%s:%i Warning jutil_name2sig_cleanup( 0x%s): "
		     "*psig is NULL\n", __FILE__, __LINE__, psig);
      return;
    }
  if ((*psig)[ 1])
    {
      free( (void*)*psig);
      *psig = NULL;
    }
}
