/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is "Java-Python Extension (JPE)".
 * 
 * The Initial Developer of the Original Code is Frederic Bruno Giacometti.
 * Portions created by Frederic Bruno Giacometti are
 * Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
 * 
 * Contributor(s): frederic.giacometti@arakne.com
 *
 * Acknowledgments:
 * Particular gratitude is expressed to the following parties for their
 * contributing support to the development of JPE:
 *     - Arakne SARL (France) and its staff (http://www.arakne.com);
 *       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
 *     - The European Union through its funding of the IST Alice project
 *     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
 *       coordinator of the IST Alice project
 *     - iPhrase Technologies Inc. (Massachusetts, USA)
 */

package python;

import java.lang.Comparable;
import java.util.List;
import java.util.Map;

/** Root class for Python objects. */

public class PyObject implements Comparable
{
  /** reference to Python None object */
  static public final PyObject None =
      new PyObject( Javapy.getNone( PyThreadState.get()));

  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( boolean val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( short val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( int val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( long val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( float val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyNumber asPython( double val)
  {
    return new PyNumber( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python string corresponding to 'val'
   */
  static public PyString asPython( String val)
  {
    return new PyString( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PySequence asPython( Object[] val)
  {
    return new PySequence( val);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static public PyObject asPython( List list)
  {
    return new PySequence( list);
  }
  /** convert to Python
   * @param val element to be converted
   * @return Python object corresponding to 'val'
   */
  static PyObject asPython( Map map)
  {
    return new PyDict( map);
  }
  /** create a new Java sibling for the Python object
   * @param val Python object
   * @return new Java sibling for the same Python object
   *   already referred to by 'val'
   */
  static public PyObject asPython( PyObject val)
  {
    return new PyObject( val.pself);
  }
  /** convert to Python
   * @param obj object to be converted
   * @return ('PyObject.asPython( PyObject val)'
   *   IF 'obj' is a PyObject instance)
   *   OR (PyObject.None IF 'obj' is 'null)
   *   OR (a Python wrapper of the java object)
   */
  static public PyObject asPython( Object obj)
  {
    PyObject pyobject = null;
    try
      {
	pyobject = (PyObject)obj;
	throw new UnknownError();
      }
    catch (ClassCastException exc)
      {
	return (null == obj) ? new PyObject( None.pself)
	    : new PyObject( Javapy.jobject_new( PyThreadState.get(), obj));
      }
    catch (UnknownError exc) // emulate try {} else {} (not in Java!?)
      {
	return new PyObject( pyobject.pself);
      }
  }

  static protected Class PyObjectClass = null;
  static protected Class MapInterface = null;
  static protected Class ListInterface = null;
  static
  {
    try
      {
	  //PyObjectClass = Class.forName( "python.PyObject");
	MapInterface = Class.forName( "java.util.Map");
	ListInterface = Class.forName( "java.util.List");
      }
    catch(ClassNotFoundException exc)
      {
	throw new IllegalArgumentException( "PyObject.class init"
					    + exc.toString());
      }
  }
  protected long pself = 0;

  private PyObject() {;}
  protected PyObject( long ptr)
  {
    pself = ptr;
    Javapy.Py_INCREF( PyThreadState.get(), pself);
  }
  
  protected void finalize()
  {
    Javapy.Py_DECREF( PyThreadState.get(), pself);
  }

  public int compareTo( Object other)
  {
    PyObject pyother = (PyObject)other;

    return pself == pyother.pself
      ? 0 : Javapy.PyObject_Compare( PyThreadState.get(),
				     pself, pyother.pself);
  }
  public boolean equals( PyObject other)
  {
    return pself == other.pself || 0 == compareTo( other);
  }
  public boolean equals( Object obj)
  {
    if (PyObjectClass == null)
      {
	try
	  {
	    PyObjectClass = Class.forName( "python.PyObject");
	  }
	catch(ClassNotFoundException exc)
	  {
	    throw new IllegalArgumentException( "PyObject.class init"
						+ exc.toString());
	  }
      }

    return PyObjectClass.isInstance( obj) ? equals( (PyObject)obj)
      : PyMapping.check( this) ? (new PyMapping( this)).equals( obj)
        : PySequence.check( this) ? (new PySequence( this)).equals( obj)
          : false;
  }
  protected Object clone()
  {
    throw new UnsupportedOperationException( "PyObject::clone()");
  }	     
  public int hashCode()
  {
    return Javapy.PyObject_Hash( PyThreadState.get(), pself);
  }
  public boolean isTrue()
  {
    return Javapy.PyObject_IsTrue( PyThreadState.get(), pself);
  }
  public PyType getPyType()
  {
    return new PyType( Javapy.PyObject_Type( PyThreadState.get(), pself));
  }
  public long getPyLength()
  {
    return Javapy.PyObject_Length( PyThreadState.get(), pself);
  }
  public Object getItem( PyObject key)
  {
    return Javapy.ptr2object(
			  Javapy.PyObject_GetItem( PyThreadState.get(), pself,
						   key.pself));
  }
  public Object getItem( Object key) { return getItem( asPython( key)); }
  public Object getItem( int index) { return getItem( asPython( index)); }
  public void setItem( PyObject key, PyObject val)
  {
    Javapy.PyObject_SetItem( PyThreadState.get(), pself, key.pself, val.pself);
  }
  public void setItem( Object key, Object val)
  {
    setItem( asPython( key), asPython( val));
  }
  public void delItem( PyObject key)
  {
    Javapy.PyObject_DelItem( PyThreadState.get(), pself, key.pself);
  }
  public String toString() { return str(); }
  public String str()
  {
    return Javapy.PyObject_Str( PyThreadState.get(), pself);
  }
  public String repr()
  {
    return Javapy.PyObject_Repr( PyThreadState.get(), pself);
  }
  public Object asObject()
  {
    // should also cover standard types....
    return Javapy.PyJobject_AsJobject( PyThreadState.get(), pself);
  }
  public boolean hasattr( String attr)
  {
    return Javapy.PyObject_HasAttrString( PyThreadState.get(), pself, attr);
  }
  public void delattr( String attr)
  {
    Javapy.PyObject_DelAttrString( PyThreadState.get(), pself, attr);
  }
  public Object getattr( String name)
  {
    return Javapy.ptr2object(
			   Javapy.PyObject_GetAttrString( PyThreadState.get(),
							  pself, name));
  }
  public void setattr( String name, PyObject obj)
  {
    Javapy.PyObject_SetAttrString( PyThreadState.get(),
				   pself, name, obj.pself);
  }
  public boolean callable()
  {
    return Javapy.PyCallable_Check( PyThreadState.get(), pself);
  }

  /** call the Python object
   * @param tuple Python tuple of arguments
   * @return result of the call
   * @throws python.PyException
   */
  public Object call( PyTuple tuple)
  {
	return Javapy.ptr2object(
			Javapy.PyObject_CallObject( PyThreadState.get(), pself,
						    tuple.pself));
  }
  /** call the Python object
   * @param args array of Python arguments
   * @return result of the call
   * @throws python.PyException
   */
  public Object call( PyObject[] args)
  {
    return Javapy.ptr2object(
			 Javapy.PyObject_CallArray( PyThreadState.get(), pself,
						    args));
  }
  /** call the Python object - todo
   * @param args array of Java arguments
   * @return result of the call
   * @throws java.lang.UnsupportedOperationException
   */
  public Object call( Object[] args)
  {
      throw new UnsupportedOperationException();
  }
  /** call the Python object - todo
   * @param args array of Java arguments
   * @param keywords keyword dictionary
   * @return result of the call
   * @throws java.lang.UnsupportedOperationException
   */
  public Object call( Object[] args, Map keywords)
  {
      throw new UnsupportedOperationException();
  }
  /** call the Python object with no argument
   * @return result of the call
   * @throws python.PyException
   */
  public Object call() { return call( new PyObject[ 0]); }
  /** @deprecated */
  public Object call( boolean arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( short arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( int arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( long arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( float arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( double arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( String arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( byte arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( char arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** @deprecated */
  public Object call( Object arg)
  {
    return call( new PyObject[] { asPython( arg) });
  }
  /** call a method of the Python object
   * @param name name of the method
   * @param tuple Python tuple of arguments
   * @return result of the call
   * @throws python.PyException
   */
  public Object callMethod( String name, PyTuple tuple)
  {
    return ((PyObject)getattr( name)).call( tuple);
  }
  /** call a method of the Python object
   * @param name name of the method
   * @param args array of Python arguments
   * @return result of the call
   * @throws python.PyException
   */
  public Object callMethod( String name, PyObject[] args)
  {
    return ((PyObject)getattr( name)).call( args);
  }
  /** call a method of the Python object
   * @param name name of the method
   * @param args array of Java arguments
   * @return result of the call
   * @throws python.PyException
   */
  public Object callMethod( String name, Object[] args)
  {
    return ((PyObject)getattr( name)).call( args);
  }
  /** call a method of the Python object with no argument
   * @param name name of the method
   * @return result of the call
   * @throws python.PyException
   */
  public Object callMethod( String name)
  {
    return callMethod( name, new PyObject[ 0]);
  }
  /** @deprecated */
  public Object callMethod( String name, boolean arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, short arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, int arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, long arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, float arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, double arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, String arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, char arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, byte arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  /** @deprecated */
  public Object callMethod( String name, Object arg)
  {
    return ((PyObject)getattr( name)).call( arg);
  }
  public void clear()
  {
    throw new UnsupportedOperationException( "PyObject::clear()");
  }
}
