/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is "Java-Python Extension (JPE)".
 * 
 * The Initial Developer of the Original Code is Frederic Bruno Giacometti.
 * Portions created by Frederic Bruno Giacometti are
 * Copyright (C) 2000-2001 Frederic Bruno Giacometti. All Rights Reserved.
 * 
 * Contributor(s): frederic.giacometti@arakne.com
 *
 * Acknowledgments:
 * Particular gratitude is expressed to the following parties for their
 * contributing support to the development of JPE:
 *     - Arakne SARL (France) and its staff (http://www.arakne.com);
 *       and in particular to Jean Hemmi, Marie-Therese Plaze and Didier Rano
 *     - The European Union through its funding of the IST Alice project
 *     - Dean Vucinic from Vrije Universiteit Brussel (Belgium),
 *       coordinator of the IST Alice project
 *     - iPhrase Technologies Inc. (Massachusetts, USA)
 */

package python;

/** Protected class - Thread-handling material. */
class PyThreadState
{
  static private ThreadLocal localThreadState = new ThreadLocal();
  static private long pyInterp = 0;
  static private boolean ownInterp = false;
  /** Turn on/off for enabling/disabling trace and debug statements. */
  static protected boolean debug = false;
  static protected boolean lock_use = false;
  static
  {
    if (debug) System.out.println( "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^");
    synchronized( PyThreadState.class)
      {
	if (Javapy.Py_IsInitialized())
	  {
	    pyInterp = 0;
	    ownInterp = false;
	    if (debug) System.out.println( "--------------------------------");
	  }
	else
	  {
	    long pySThread;
	    pySThread = Javapy.Py_Initialize();
	    pyInterp = Javapy.interpFromThread( pySThread);
	    ownInterp = true;
	    Javapy.attach( pySThread, true);
	  }
      }
    if (debug) System.out.println( "++++++++++++++++++++++++++++++++++++");
  }
  static synchronized protected void classFinalize()
  {
    if (debug) System.out.println( "|| PyThreadState.classFinalize()");
    if (ownInterp)
      {
	Javapy.Py_Finalize();
	ownInterp = false;
	pyInterp = 0;
      }
  }

  protected long threadState = 0;
  protected boolean ownThread = false;
  protected boolean locked = false;

  public String toString()
  {
    return "(0x" + Long.toHexString( threadState)
      + " ownThread=" + (ownThread ? "1" : "0")
      + " locked=" + (locked  ? "1" : "0") + ")";
  }

  private PyThreadState( long vthreadState, boolean owned)
  {
    if (debug) System.out.println( ">>> PyThreadState( 0x"
				   + Long.toHexString( vthreadState)
				   + ", " + (owned ? "1" : "0")
				   + ")");
    if (pyInterp == 0)
      {
	pyInterp = Javapy.interpFromThread( vthreadState);
      }
    else if (pyInterp != Javapy.interpFromThread( vthreadState))
      {
	throw new IllegalStateException(
					"Java threads cannot be associated "
					+ "to Python threads attached to "
					+ "different Python interpreters");
      }
    threadState = vthreadState;
    ownThread = owned;
    if (debug) System.out.println( "<<< PyThreadState( 0x"
				   + Long.toHexString( vthreadState)
				   + ", " + (owned ? "1" : "0")
				   + ")");
  }

  protected void finalize()
  {
    if (debug) System.out.println( ">>> PyThreadState.finalize() "
				   + this.toString());
    if (ownThread) Javapy.threadStateDelete( threadState);
  }

  static synchronized private void attach( long lThreadState, boolean owned)
    // python callback from Javapy.attach()
    // DO NOT call this method directly from Java
  {
    if (0 == lThreadState)
      throw new NullPointerException( "Null PyThreadState");
    if (null == localThreadState.get())
      {
	if (debug)
	  System.out.println( "PyThreadState.attach(): "
			      + "localThreadState.set( );");
	localThreadState.set( new PyThreadState( lThreadState, owned));
      }
    else
      {
	PyThreadState myState = (PyThreadState)localThreadState.get();
	if (lThreadState == myState.threadState)
	  {
	    System.err.println( "Warning: python.PyThreadState.attach() - "
				+ "Java thread already attached "
				+ "to the same Python thread"
				+ myState.toString());
	  }
	else
	  {
	    System.err.println( "Error: python.PyThreadState.attach() - "
				+ "Java thread already attached "
				+ "to a different Python thread"
				+ myState.toString() + " && 0x"
				+ Long.toHexString( lThreadState));
	    throw new IllegalStateException( localThreadState.get().toString()
					     + " && 0x"
					     + Long.toHexString(lThreadState));
	  }
      }
  }

  static protected PyThreadState pyThread()
  {
    if (debug) System.out.println( ">>> PyThreadState.pyThread()");
    PyThreadState result = (PyThreadState)localThreadState.get();
    if (result == null)
      { //create owned Python thread
	if (debug) System.out.println( "||| PyThreadState.pyThread(): "
				       + "set new PyThreadState()");
	if (pyInterp == 0)
	  {
	    if (debug) System.err.println( "PyThreadState.java: ERROR - "
					   + "no Python interpreter attached");
	    throw new NullPointerException( "no Python interpreter attached");
	  }
	Javapy.attach( Javapy.threadStateNew( pyInterp), true);
	result = (PyThreadState)localThreadState.get();
	if (result == null)
	  {
	    if (debug) System.err.println( "PyThreadState.java: ERROR - "
					   + "still no attachment");
	    throw new NullPointerException( "still no attachment");
	  }
      }
    if (debug) System.out.println( "<<< PyThreadState.pyThread() "
				   + result.toString());
    return result;
  }

  static protected long get()
  {
    PyThreadState thr = pyThread();
    if (lock_use) return thr.locked ? 0 : thr.threadState;
    else return thr.threadState;
  }

  synchronized protected void lock()
  {
    if (!lock_use)
      throw new IllegalStateException( "not lock_use");
    if (debug)
      {
	System.out.println( ">>> PyThreadState.lock()");
	if (0 == threadState)
	  throw new NullPointerException( "0 == threadState");
      }
    if (locked) throw new RuntimeException( "lock must not be held");
    Javapy.PyEval_AcquireThread( threadState);
    locked = true;
    if (debug) System.out.println( "<<< PyThreadState.lock()");
  }

  synchronized protected void unlock()
  {
    if (!lock_use)
      throw new IllegalStateException( "not lock_use");
    if (debug) System.out.println( ">>> PyThreadState.unlock()");
    if (!locked) throw new RuntimeException( "lock must be held");
    Javapy.PyEval_SaveThread();
    locked = false;
    if (debug) System.out.println( "<<< PyThreadState.unlock()");
  }
}
